################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#       You can find more information at http://gato.sf.net
#
#	file:   AnimatedAlgorithms.py
#	author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################
import GatoGlobals
from DataStructures import VertexLabeling, Queue
from AnimatedDataStructures import *
#from GraphDisplay import GraphDisplay
#from Graph import SubGraph

g = GatoGlobals.AnimationParameters

def shortestPath(G, A, s, t):
    """ Find a shortest path and return it as a set of edges. If no
        path exists, it returns None """
    pred = AnimatedVertexLabeling(A)    
    Q = AnimatedVertexQueue(A)    
    A.SetAllEdgesColor(g.cEdgeDefault)
    A.SetAllVerticesColor(g.cVertexDefault)
        
    for v in G.vertices:
        pred[v] = None	
    Q.Append(s)
    
    while Q.IsNotEmpty() and pred[t] == None:
        v = Q.Top()
        for w in AnimatedNeighborhood(A,G,v):
            if pred[w] == None and w != s:
                pred[w] = v
                Q.Append(w)
                
    if pred[t] == None: # No augmenting path found
        return None
        
    path = []
    v = t
    while pred[v] != None:
        A.SetVertexColor(v,"red")
        A.SetEdgeColor(pred[v],v,"red")
        path.append((pred[v],v))
        v = pred[v]
    A.SetVertexColor(v,"red")
    return path
    



################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#	file:   ObjectGraph.py
#	author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 2007-2015 Alexander Schliep, Winfried Hochstaettler
#                                   
#       Contact: alexander@schliep.org             
#
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################

from GatoGlobals import *
from DataStructures import Point2D, VertexLabeling, EdgeLabeling, EdgeWeight
#from math import log

import logging
log = logging.getLogger("ObjectGraph.py")

################################################################################
#
# Base classes
#
################################################################################

class VertexObject(object):
     # Python 2.3 kludge
    def nrOfVertexWeights():
        return 0
    NrOfVertexWeights = staticmethod(nrOfVertexWeights)

    def __init__(self):
        self.id = None
        self.embedding = None
        self.labeling = None
        self.inEdges = []
        self.outEdges = []
           


class EdgeObject(object):
    # Python 2.3 kludge
    def nrOfEdgeWeights():
        return 2
    NrOfEdgeWeights = staticmethod(nrOfEdgeWeights)

    def __init__(self, tail, head):
        self.tail = tail
        self.head = head
        self.weight = 0.0

    def key(self):
        return (self.tail.id, self.head.id)

    def GetEdgeWeight(self,i):
        if i == 0:
            return self.weight

    def SetEdgeWeight(self,i,value):
        if i == 0:
            self.weight = value
        

        


################################################################################
#
# ObjectGraph
#
################################################################################
class ObjectGraph(object):
    """ 

    
    """
    
    def __init__(self, vertexClass, edgeClass):
        self.vertexClass = vertexClass
        self.edgeClass = edgeClass
        self.vertices = {}
        self.edges = {}
        self.highVertexID     = 0    # INTERNAL
        self.simple           = 1
        self.euclidian        = 1
        self.directed         = 0


        self.edgeWeights      = {}   # Dictionary of edge labellings
        self.edgeWeights[0]   = EdgeWeight(self)
        self.vertexWeights    = {}   # None by default
        self.size             = 0
        self.edgeWidth        = None
        self.vertexAnnotation = None
        self.edgeAnnotation   = None
        self.properties       = {}
        
    def AddVertex(self):
        """ Add an isolated vertex. Returns the id of the new vertex """
        v = self.vertexClass()
        v.id = self.GetNextVertexID()
        self.vertices[v.id] = v
        return v.id
        
    def DeleteVertex(self, v):
        """ Delete the vertex v and its incident edges """
        outEdges = self.vertices[v].outEdges[:] # Need a copy here
        inEdges = self.vertices[v].inEdges[:]
        for e in outEdges:
            (v,w) = e.key()
            self.DeleteEdge(v,w)
        for e in inEdges:
            (w,v) = e.key()            
            if w != v: # We have already deleted loops
                self.DeleteEdge(w,v)
        del(self.vertices[v])
        
    def QVertex(self, v):
        """ Check whether v is a vertex """
        return v in self.vertices.keys()

        
    def AddEdge(self,tail,head):
        """ Add an edge (tail,head). Returns nothing
            Raises GraphNotSimpleError if
            - trying to add a loop
            - trying to add an edge multiply 
        
            In case of directed graphs (tail,head) and (head,tail)
            are distinct edges """
        
        if self.simple == 1 and tail == head: # Loop
            raise GraphNotSimpleError, '(%d,%d) is a loop' % (tail,head)
        if self.directed == 0 and self.edges.has_key((head,tail)):
            raise GraphNotSimpleError, '(%d,%d) is already an undirected edge' % (head,tail)
        if self.edges.has_key((tail,head)): # Multiple edge
            raise GraphNotSimpleError, '(%d,%d) is already an directed edge' % (tail,head)

        e = self.edgeClass(self.vertices[tail],self.vertices[head])
        self.edges[(tail,head)] = e
        self.vertices[tail].outEdges.append(e)
        self.vertices[head].inEdges.append(e)

        if self.QEuclidian():
            t = self.GetEmbedding(tail)
            h = self.GetEmbedding(head)
            self.SetEdgeWeight(0,tail,head,sqrt((h.x - t.x)**2 + (h.y - t.y)**2))
        else:
            self.SetEdgeWeight(0,tail,head,0)
        for i in xrange(1,self.NrOfEdgeWeights()):
            self.SetEdgeWeight(i,tail,head,0)

        
    def DeleteEdge(self,tail,head):
        """ Deletes edge (tail,head). Does *not* handle undirected graphs
            implicitely. Raises NoSuchEdgeError upon error. """

        try:
            e = self.edges[(tail,head)]
        except KeyError:
            raise NoSuchEdgeError, "(%d,%d) is not an edge." % (tail,head)

        self.vertices[tail].outEdges.remove(e)
        self.vertices[head].inEdges.remove(e)
        del(self.edges[(tail,head)])
            
            
    def Edge(self,tail,head):
        """  Handles undirected graphs by returning correctly ordered
             vertices as (tail,head). Raises NoSuchEdgeError upon error. """
        
        if tail not in self.vertices.keys() or head not in self.vertices.keys():
            raise NoSuchEdgeError, "(%d,%d) is not an edge." % (tail,head)
            
        if self.edges.has_key((tail,head)):
            return (tail,head)
        elif self.directed == 0 and self.edges.has_key((head,tail)):
            return (head,tail)
        else:
            raise NoSuchEdgeError, "(%d,%d) is not an edge." % (tail,head)
            
            
    def QEdge(self,tail,head):
        """ Returns 1 if (tail,head) is an edge in G. If G is undirected
            order of vertices does not matter """
        if self.directed == 1:	
            return self.edges.has_key((tail,head))
        else: 
            return self.edges.has_key((tail,head)) or self.edges.has_key((head,tail))


    def QEdgeWidth(self):
        """ Returns 1 if individual edge widths are defined, 0 else """
        return 0 # XXX NOT SUPPORTED
        

    def EdgeWidth(self, tail, head):
        return 0 # XXX NOT SUPPORTED 

            
    def Neighborhood(self,v):
        """ Returns the vertices which are connected to v. Does handle
            undirected graphs (i.e., returns vertices w s.t. either 
            (v,w) or (w,v) is an edge) """
        
        if self.directed:
            return self.OutNeighbors(v)
        else:
            return self.InOutNeighbors(v)
            
            
    def InNeighbors(self,v):
        """ Returns vertices w for which (w,v) is an edge """
        return map(lambda e: e.tail.id, self.vertices[v].inEdges)
        
        
    def OutNeighbors(self,v):
        """ Returns vertices w for which (v,w) is an edge """
        return map(lambda e: e.head.id, self.vertices[v].outEdges)
        
        
    def InOutNeighbors(self,v):
        """ Returns vertices w for which (v,w) or (w,v) is an edge """	
        return self.InNeighbors(v) + self.OutNeighbors(v)
        
        
    def InEdges(self,v):
        """ Returns edges (*,v) """	
        return map(lambda e: e.key(),self.vertices[v].inEdges) 
        
        
    def OutEdges(self,v):
        """ Returns edges (v,*) """	
        return map(lambda e: e.key(),self.vertices[v].outEdges) 
        
        
    def IncidentEdges(self,v):
        """ Returns edges (v,*) and (*,v) """	
        return self.InEdges(v) + self.OutEdges(v)
        
        
    def Edges(self):
        """ Returns all edges """		
        return self.edges.keys()

        
    def Vertices(self):
        """ Returns all edges """		
        return self.vertices.keys()
        
    def printMy(self):
        """ Debugging only """
        for v in self.vertices:
            print v, " -- ", self.adjLists[v]
            
            
    def GetNextVertexID(self):
        """ *Internal* returns next free vertex id """
        self.highVertexID += 1
        return self.highVertexID
        
        
    def Order(self):
        """ Returns order i.e., the number of vertices """
        return len(self.vertices.keys())
        
        
    def Size(self):
        """ Returns size i.e., the number of edge """
        return len(self.edges.keys()) 
        
        
    def Degree(self, v):
        """ Returns the degree of the vertex v, which is
            - the number of incident edges in the undirect case
            - the number of outgoing edges in the directed case """
        
        if self.directed:
            return len(self.vertices[v].outEdges)
        else:
            return len(self.vertices[v].outEdges) + len(self.vertices[v].inEdges)

            
    def InDegree(self, v):
        """ Returns the number of incoming edges for direct graphs """
        if self.directed:
            return len(self.vertices[v].inEdges)
        else:
            return None # Proper error to raise?
            
            
    def OutDegree(self, v):
        """ Returns the number of incoming edges for direct graphs """
        if self.directed:
            return len(self.vertices[v].outEdges)
        else:
            return None # Proper error to raise?
            
            
    def QEuclidian(self):
        """ Returns 1 if the graph is euclidian, 0 else """
        return self.euclidian
        
        
    def QDirected(self):
        """ Returns 1 if the graph is directed, 0 else """
        return self.directed
        
        
    def CalculateWidthFromWeight(self, scale, weightID = 0):
        """ Calculate width of edges (self.edgeWidth will be used by 
            GraphDisplay if not none) from the specified set of edge
            weights. 
        
            Default: weightID = 0 is used """
        pass  # XXX NOT SUPPORTED
##        self.edgeWidth = EdgeLabeling()
##        edges = self.Edges()
##        maxWeight = max(self.edgeWeights[weightID].label.values())
##        for e in edges:
##            self.edgeWidth[e] = scale * (1 + 35 * self.edgeWeights[weightID][e] / maxWeight) 
            
    def NrOfEdgeWeights(self):
        return self.edgeClass.NrOfEdgeWeights()

    def SetEdgeWeight(self,i,v,w,value):
        self.edges[(v,w)].SetEdgeWeight(i,value)

    def GetEdgeWeight(self,i,v,w):
        return self.edges[(v,w)].GetEdgeWeight(i)

    def QIntegerWeight(self,i):
        # Edge and Vertex class should take care of that ...class method
        return 0 # XXX UNSUPPORTED: FLOAT WEIGHTS ONLY. Barf
    

    def NrOfVertexWeights(self):
        return self.vertexClass.NrOfVertexWeights()

    def SetVertexWeight(self,i,v,value):
        pass 

    def GetVertexWeight(self,i,v):
        pass

    def GetLabeling(self,v):
        return self.vertices[v].labeling
    
    def SetLabeling(self,v, value):
        self.vertices[v].labeling = value

    def GetEmbedding(self,v):
        return self.vertices[v].embedding
    
    def SetEmbedding(self,v, x, y):
        self.vertices[v].embedding = Point2D(x,y)

                  
    def Euclidify(self):
        """ Replace edge weights with weightID = 0 with Euclidean distance 
            between incident vertices """
        pass  # XXX NOT SUPPORTED 
##        for v in self.vertices:
##            for w in self.adjLists[v]:
##                d = ((self.embedding[v].x - self.embedding[w].x)**2 + 
##                     (self.embedding[v].y - self.embedding[w].y)**2)**(.5)
                
##                if self.edgeWeights[0].QInteger():
##                    self.edgeWeights[0][(v,w)] = int(round(d))
##                else:
##                    self.edgeWeights[0][(v,w)] = d
                    
##        self.euclidian = 1
        
        
    def Integerize(self, weightID = 0):
        """ Integerize: Make all edge weights integers """
        pass  # XXX NOT SUPPORTED 
##        if weightID == 'all':
##            for w in self.edgeWeights.keys():
##                self.edgeWeights[w].Integerize()
##        else:
##            self.edgeWeights[weightID].Integerize()
            
            
    def Undirect(self):
        """ If (u,v) and (v,u) are edges in the directed graph, remove one of them.
            to make graph undirected (no multiple edges allowed). Which one gets
            deleted depends on ordering in adjacency lists. """
        if not self.directed:
            return
            
        for v in self.vertices.keys():
            for e in self.vertices[v].outEdges:
                w = e.head.id
                if v in self.OutNeighbors(w):
                    self.DeleteEdge(w,v)
                    
        self.directed = 0
        
    def SetProperty(self, name, val):
        """ Set the value of property 'name' to 'val' """
        pass # XXX NOT SUPPORTED 
        #self.properties[name] = val
        
    def Property(self,name):
        """ Return the value of property 'name'. If the property
           'name' has not been set 'Unknown' is returned """
        pass  # XXX NOT SUPPORTED 
        #try:
        #    return self.properties[name]
        #except:
        #    return 'Unknown'
            
    def About(self):
        """ Return string containing HTML code providing information
            about the graph """
        return "<HTML><BODY> <H3>No information available</H3></BODY></HTML>"
        
        
##        ################################################################################
##        #
##        # Induced Subgraph
##        #
##        ################################################################################
        
##class SubGraph(Graph):
##    """ Provides a subgraph, i.e., a subset of the vertices and edges 
##        of a specified graph
    
##        Vertices are specified via ids from its supergraph and edges via
##        (tail,head)-tuples 
    
##        It also keeps track of the subgraphs total weight (= sum of edge 
##        weights) for weights with weightID == 0
##    """
    
    
##    def __init__(self,G):
##        Graph.__init__(self)
##        self.superGraph    = G
        
##        self.embedding     = self.superGraph.embedding
##        self.labeling      = self.superGraph.labeling
##        self.edgeWeights   = self.superGraph.edgeWeights
        
##        self.directed = self.superGraph.directed
        
##        self.totalWeight   = 0
        
        
##    def AddVertex(self,v):
##        """ Add a vertex from the supergraph to the subgraph.
##            Returns NoSuchVertexError if v does not exist in
##            supergraph """
##        try:
##            self.vertices.append(v)
##            #f = lambda x, vertexList=self.vertices: x in vertexList
##            #self.adjLists[v]    = filter(f, self.superGraph.adjLists[v])
##            #self.invAdjLists[v] = filter(f, self.superGraph.invAdjLists[v])
##            self.adjLists[v]    = []
##            self.invAdjLists[v] = []
##        except:
##            raise NoSuchVertexError, "%d is not a vertex in the supergraph" % v
            
##    def AddEdge(self,tail,head):
##        """ Add an edge from the supergraph to the subgraph.
##            Will also add tail and/or head if there are not
##            already in subgraph """
##        try:
##            if not tail in self.vertices:
##                self.AddVertex(tail)
##            if not head in self.vertices:
##                self.AddVertex(head)
##            (tail,head) = self.superGraph.Edge(tail,head) 
            
##            self.adjLists[tail].append(head)
##            self.invAdjLists[head].append(tail)
##            self.size = self.size + 1
##            try:
##                w = self.superGraph.edgeWeights[0][(tail,head)]
##            except KeyError:
##                w = 0.0 # XXX we dont have w weight for the edge. Make totalWeight configurable/subclass
##            self.totalWeight += w
            
##        except (KeyError, NoSuchVertexError, NoSuchEdgeError):
##            raise NoSuchEdgeError, "(%d,%d) is not an edge in the supergraph." % (tail,head)
            
##    def AddSubGraph(self,G):
##        """ Add subgraph G to self. Will do nothing if self and G 
##            have distinct supergraphs """
##        if self.superGraph != G.superGraph:
##            log.error("AddSubGraph: distinct superGraphs")
##            return
##        for v in G.vertices:
##            self.AddVertex(v)
##        for e in G.Edges():
##            self.AddEdge(e[0],e[1])
            
            
##    def DeleteEdge(self,tail,head):
##        """ Delete edge from subgraph. Raises NoSuchEdgeError
##            upon error """
##        if tail in self.vertices and head in self.vertices:
##            superEdge = self.superGraph.Edge(tail,head)
##            self.totalWeight =  self.totalWeight - self.superGraph.edgeWeights[0][superEdge]
##            self.adjLists[tail].remove(head)
##            self.invAdjLists[head].remove(tail)
##            self.size = self.size - 1
##        else:
##            raise NoSuchEdgeError, "(%d,%d) is not an edge." % (tail,head)
            
##    def Clear(self):
##        """ Delete all vertices and edges from the subgraph. """
##        self.vertices         = [] 
##        self.adjLists         = {}
##        self.invAdjLists      = {}   # Inverse Adjazenzlisten
##        self.size = 0
##        self.totalWeight   = 0
        
        
##    def GetNextVertexID(self):
##        """ *Internal* safeguard """
##        log.error("Induced Subgraph -> GetNextVertexID should never have been called")
        
##    def Weight(self):
##        """ Returns the total weight (= sum of edge weights) of subgraph """
##        return self.totalWeight
        
        
##    def QEuclidian(self):
##        """ Returns 1 if the super graph is euclidian, 0 else """
##        return self.superGraph.euclidian
        
        
##    def QDirected(self):
##        """ Returns 1 if the super graph is directed, 0 else"""
##        return self.superGraph.directed
        
##    def QEdge(self,tail,head):
##        """ Returns 1 if (tail,head) is an edge in G """
##        if not tail in self.vertices or not head in self.vertices:
##            return 0
##        if self.directed == 1:	
##            return head in self.adjLists[tail]
##        else: 
##            return head in self.adjLists[tail] or tail in self.adjLists[head]
            
            



#!/usr/bin/env python2.6
################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#	file:   GatoFile.py
#	author: Achim Gaedke (achim.gaedke@zpr.uni-koeln.de)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################

import Tkinter

class dom_structure_widget(Tkinter.Frame):
    """
    this widget displays the structure of a dom tree
    """
    
    class nodeDisplayProperties:
        """
        holds all properties for node
        """
        def __init__(self):
            self.expanded=0
            self.fg_color=""
            self.bg_color=""
            self.tag=""
            
    def __init__(self,master,
                 dom,
                 report_function=None,
                 cnf={},
                 **config):
        """
        creates the dom tree widget
        """
        
        # open a text widget
        my_defaults={"highlightthickness":"0"}
        my_defaults.update(config)
        my_defaults.update(cnf)
        if my_defaults.has_key("font"): del my_defaults["font"] 
        text_defaults={'bg':'white','cursor':'top_left_arrow',
                       'selectbackground':"white", 'selectborderwidth':'0',
                       "insertwidth":"0", "exportselection":"0",
                       "highlightthickness":"0"}
        text_defaults.update(config)
        text_defaults.update(cnf)
        Tkinter.Frame.__init__(self,master,cnf=my_defaults)
        self.textWidget=Tkinter.Text(self,cnf=text_defaults)
        self.textWidget.pack(side=Tkinter.LEFT,fill=Tkinter.BOTH,expand=1)
        self.scroll=Tkinter.Scrollbar(self,command=self.textWidget.yview)
        self.textWidget.configure(yscrollcommand=self.scroll.set)
        self.scroll.pack(side=Tkinter.LEFT,fill=Tkinter.Y,expand=1)
        self.dom=dom
        self.selectedNode=None
        self.report_function=report_function
        self.indent_step="    "
        self.expand_sign="+"
        self.collapse_sign="-"
        
        #parameters for default isVisible function
        self.mergeText      = 0 # if contiguous text nodes should be merged
        self.textTypes      = [xml.dom.Node.TEXT_NODE, xml.dom.Node.CDATA_SECTION_NODE]
        self.ignorableTypes = self.textTypes # ignore text by default
        
        # without normal state no key events
        self.textWidget.configure(state=Tkinter.NORMAL)
        # start with root element
        self.textWidget.mark_set(Tkinter.INSERT,Tkinter.END)
        self.textWidget.mark_gravity(Tkinter.INSERT,Tkinter.RIGHT)
        self.createNodeEntries("root",dom,"",Tkinter.INSERT,-1)
        
        # enable node selection
        self.selectedTag=None
        self.textWidget.tag_bind("collapse","<Button-1>",self.collapseNodeEvent)
        self.textWidget.tag_bind("expand","<Button-1>",self.expandNodeEvent)
        self.textWidget.tag_bind("node","<Button-1>",self.selectNodeEvent)
        self.textWidget.tag_bind("node","<Double-Button-1>",self.chooseNodeEvent)
        self.textWidget.bind("<Key>",lambda e:"break")
        self.textWidget.bind("<Up>"  , self.prevNodeEvent)
        self.textWidget.bind("<Down>", self.nextNodeEvent)
        self.textWidget.bind("<Left>"  , self.collapseNodeEvent)
        self.textWidget.bind("<Right>", self.expandNodeEvent)
        self.textWidget.bind("<Return>",self.chooseNodeEvent)
        
    def isVisible(self, node):
        """
        default function for visibility of nodes
        """
        # simple case: ignore all types of this case
        if node.nodeType in self.ignorableTypes:
            return 0
            
            # merge subsequent text nodes and display only first.
        if self.mergeText and node.nodeType in self.textTypes:
            lastNode=node.previousSibling
            if lastNode:
                return lastNode.nodeType in self.textTypes
            else:
                return 1
                
        return 1
        
    def nameNode(self, node):
        return xmlDecode(node.nodeName)
        
    def setNodeColor(self, node, fg="", bg=""):
        """
        sets the color of a node
        """
        if not node.__dict__.has_key("DisplayProperties"):
            node.DisplayProperties=dom_structure_widget.nodeDisplayProperties()
        if fg:
            node.DisplayProperties.fg_color=fg
        if bg:
            node.DisplayProperties.bg_color=bg
        if node.DisplayProperties.tag:
            self.textWidget.tag_config("nodeName-"+node.DisplayProperties.tag,
                                       foreground=node.DisplayProperties.fg_color,
                                       background=node.DisplayProperties.bg_color)
            
    def createNodeEntries(self,subtag,subtree,indentation,mark_name,depth=-1):
        """
        recursive creation of all node entries
        """
        if depth==0: return
        i=0 # counter of nodes to identifiy the label
        if not subtree.__dict__.has_key("DisplayProperties"):
            subtree.DisplayProperties=dom_structure_widget.nodeDisplayProperties()
        subtree.DisplayProperties.expanded=1
        
        for node in subtree.childNodes:
        
            # select nodes that are displayed
            if not self.isVisible(node):
                # do not display it
                i+=1
                continue
                
            this_subtag="%s-%d"%(subtag,i)
            # node will be displayed, so add property tag
            if not node.__dict__.has_key("DisplayProperties"):
                node.DisplayProperties=dom_structure_widget.nodeDisplayProperties()
            node.DisplayProperties.tag=this_subtag
            self.textWidget.tag_config("nodeName-"+this_subtag,
                                       foreground=node.DisplayProperties.fg_color,
                                       background=node.DisplayProperties.bg_color)
            
            # has this node visible children ? -> is it expandable ?
            if node.hasChildNodes() and filter(self.isVisible,node.childNodes):
                # yes, it has children
            
                # decide if node is expanded or not
                if depth==1:
                    node.DisplayProperties.expanded=0
                elif depth>1:
                    node.DisplayProperties.expanded=1
                    # otherwise, do it as done last time
                    
                    # indent
                self.textWidget.insert(mark_name, indentation, this_subtag)
                # draw node
                if node.DisplayProperties.expanded:
                    self.textWidget.insert(mark_name,self.collapse_sign,(this_subtag,"collapse"))
                else:
                    self.textWidget.insert(mark_name,self.expand_sign,(this_subtag,"expand"))
                    # write name
                self.textWidget.insert(mark_name,
                                       self.nameNode(node),
                                       (this_subtag,"node","nodeName-"+this_subtag))
                # write end of line
                self.textWidget.insert(mark_name, "\n", this_subtag)
                if node.DisplayProperties.expanded and depth!=1:
                    # one more level to write
                    self.createNodeEntries(this_subtag,
                                           node,
                                           indentation+self.indent_step,
                                           mark_name,
                                           depth-1)
            else:
                # no, no children
                # write indentation and no collapse/expand handle
                self.textWidget.insert(mark_name, indentation+" ", this_subtag)
                # now name
                self.textWidget.insert(mark_name,
                            self.nameNode(node),
                            (this_subtag,"node","nodeName-"+this_subtag))
                # now end of line
                self.textWidget.insert(mark_name, "\n", this_subtag)
            i+=1
            
    def collapseNodeEvent(self,event):
        """
        find tag, that denotes the subtree and pass to collapse_tag
        """
        if event.type=="4":
            # collapse sign hit by mouse
            my_tags=self.textWidget.tag_names("@%d,%d"%(event.x,event.y))
            tree_tags=filter(lambda t:t[:4]=='root',my_tags)
            if len(tree_tags)!=1:
                print "not good: found ",tree_tags,"to collapse!"
                return "break"
            self.collapse_tag(tree_tags[0])
            return "break"
        elif event.type=="2":
            # left arrow key, ToDo
            return "break"
        else:
            return
            
    def expandNodeEvent(self,event):
        """
        find tag, that denotes the subtree and pass to collapse_tag
        """
        if event.type=="4":
            # collapse sign hit by mouse
            my_tags=self.textWidget.tag_names("@%d,%d"%(event.x,event.y))
            tree_tag=filter(lambda t:t[:4]=='root',my_tags)
            if len(tree_tag)!=1:
                print "not good: found ",tree_tag," to expand!"
                return "break"
            self.expand_tag(tree_tag[0])
            return "break"
        elif event.type=="2":
            # right arrow key, ToDo
            return "break"
        else:
            return
            
    def selectNodeEvent(self,event):
        """
        clicked once on the node
        """
        my_tags=self.textWidget.tag_names("@%d,%d"%(event.x,event.y))
        tree_tags=filter(lambda t:t[:4]=='root',my_tags)
        if len(tree_tags)!=1:
            print "not good: found ",tree_tags," to select!"
        subtree=self.subtree_from_tag(tree_tags[0],self.dom)
        # mark the node
        self.selectNode(tree_tags[0])
        return "break"
        
    def nextNodeEvent(self,event):
        """
        event for down key
        """
        if self.selectedNode is None: return "break"
        (start,end)=self.textWidget.tag_ranges(self.selectedNode.DisplayProperties.tag)
        result=self.textWidget.tag_nextrange("node",end)
        if len(result)==0: return "break"
        my_tags=self.textWidget.tag_names(result[0])
        tree_tags=filter(lambda t:t[:4]=='root',my_tags)
        if len(tree_tags)!=1:
            print "not good: found ",tree_tags," to select!"
            # mark the node
        self.selectNode(tree_tags[0])
        return "break"
        
    def prevNodeEvent(self,event):
        """
        event for up key
        """
        if self.selectedNode is None: return "break"
        (start,end)=self.textWidget.tag_ranges(self.selectedNode.DisplayProperties.tag)
        result=self.textWidget.tag_prevrange("node",start)
        if len(result)==0: return "break"
        my_tags=self.textWidget.tag_names(result[0])
        tree_tags=filter(lambda t:t[:4]=='root',my_tags)
        if len(tree_tags)!=1:
            print "not good: found ",tree_tags," to select!"
            # mark the node
        self.selectNode(tree_tags[0])
        return "break"
        
    def selectNode(self,tree_tag):
        """
        mark node as selected
        the node is specified by the tree tag
        """
        if self.selectedNode:
            self.setNodeColor(self.selectedNode,bg="white")
            # find node element for this line
            
        subtree=self.subtree_from_tag(tree_tag,self.dom)
        self.setNodeColor(subtree,bg="green")
        self.selectedNode=subtree
        
        if self.report_function and subtree:
            self.report_function("selectedNode",subtree)
            
            
    def chooseNodeEvent(self,event):
        """
        choose node: i.e. report to report function
        """
        subtree=None
        if event.type=="4":
            # selection by mouse
            my_tags=self.textWidget.tag_names("@%d,%d"%(event.x,event.y))
            tree_tags=filter(lambda t:t[:4]=='root',my_tags)
            if len(tree_tags)!=1:
                print "not good: found ",tree_tags," to select!"
                # select this node
            self.selectNode(tree_tags[0])
            subtree=self.subtree_from_tag(tree_tags[0],self.dom)
        elif event.type=="2":
            # selection by key
            if self.selectedNode is None:
                return "break"
            subtree=self.selectedNode
        else:
            # unknown
            return "break"
        if self.report_function and subtree:
            self.report_function("chosenNode",subtree)
        return "break"
        
    def collapse_tag(self,tag):
        """
        collapses the tree under the given tag
        """
        subtree=self.subtree_from_tag(tag,self.dom)
        subtree.DisplayProperties.expanded=0
        all_tags=self.textWidget.tag_names()
        # determine tags to delete
        tags_to_delete=filter(lambda s,t=tag+"-",l=len(tag)+1:s[:l]==t,all_tags)
        
        # handle selected node, that may become invisible
        if self.selectedNode is not None and self.selectedNode.DisplayProperties.tag in tags_to_delete:
            self.selectNode(tag)
            
        for t in tags_to_delete:
            self.textWidget.delete(t+".first",t+".last")
            self.textWidget.tag_delete(t)
            
            # change collapse symbol to expand symbol
        symbol_index=self.textWidget.tag_nextrange('collapse',tag+".first",tag+".last")
        self.textWidget.delete(symbol_index[0],symbol_index[1])
        self.textWidget.insert(symbol_index[0],self.expand_sign,('expand',tag))
        
    def expand_tag(self,tag):
        """
        expand the tree under the tag
        """
        subtree=self.subtree_from_tag(tag,self.dom)
        start_index=self.textWidget.index(tag+".last")
        self.textWidget.mark_set(Tkinter.INSERT,start_index)
        
        # change expand symbol to collapse symbol
        symbol_index=self.textWidget.tag_nextrange('expand',tag+".first",tag+".last")
        symbol_index=self.textWidget.tag_nextrange('expand',tag+".first",tag+".last")
        self.textWidget.delete(symbol_index[0],symbol_index[1])
        self.textWidget.insert(symbol_index[0],self.collapse_sign,('collapse',tag))
        
        # determine indentation
        indentation=self.textWidget.get(tag+".first",symbol_index[0])
        self.createNodeEntries(tag,
                               subtree,
                               indentation+self.indent_step,
                               Tkinter.INSERT,
                               -1)
        
    def subtree_from_tag(self,tag,dom):
        """
        parse tag and traverse tree
        """
        subtree=dom
        for token in tag.split('-'):
            if token=='root':
                subtree=dom
            else:
                index_nr=int(token)
                subtree=subtree.childNodes[index_nr]
        return subtree
        



#!/usr/bin/env python
################################################################################
#
#       This file is part of Gato (Graph Algorithm Toolbox) 
#
#	file:   ProbEditorWidgets.py
#	author: Achim Gaedke (achim.gaedke@zpr.uni-koeln.de)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################

import math
import Tkinter
import ProbEditorBasics

class scroll_canvas(Tkinter.Canvas):
    """
    should avoid resizing by adding or removing scrollbars in x or y-direction or both
    """
    
    def __init__(self,master,cnf={},**kw):
                #hidden frame
        self.hidden_frame=Tkinter.Frame(master)
        self.hidden_frame.rowconfigure(0,weight=1)
        self.hidden_frame.columnconfigure(0,weight=1)
        
        cnf=Tkinter._cnfmerge((cnf,kw))
        if not cnf.has_key('highlightthickness'):
            cnf['highlightthickness']=0
        Tkinter.Canvas.__init__(self,self.hidden_frame,cnf)
        
        Tkinter.Canvas.grid(self,row=0,column=0,sticky=Tkinter.NSEW)
        self.hidden_frame.bind('<Configure>',self.config_event)
        # prepare scrollbars
        self.sb_y=Tkinter.Scrollbar(self.hidden_frame,
                                    orient=Tkinter.VERTICAL,
                                    command=self.yview)
        self.sb_x=Tkinter.Scrollbar(self.hidden_frame,
                                    orient=Tkinter.HORIZONTAL,
                                    command=self.xview)
        Tkinter.Canvas.configure(self,
                                 yscrollcommand=self.sb_y.set,
                                 xscrollcommand=self.sb_x.set)
        
    def config_event(self,event):
        # returns empty list, if no scrollregion exists
        scrollr=Tkinter.Canvas.cget(self,'scrollregion')
        if scrollr:
            # print "from config",scrollr
            scrollr=map(int,scrollr.split())
        else:
            # print "from bbox"
            scrollr=self.bbox(Tkinter.ALL)
            self.configure(scrollregion=scrollr)
            ##        print "event: widht %d, height %d"%(event.width,event.height)
            ##        print "scrollregion: ",scrollr
            ##        print "winfo: widht %d, height %d"%(self.c.winfo_width(),
            ##                                            self.c.winfo_height())
            ##        print "winfo req: widht %d, height %d"%(self.c.winfo_reqwidth(),
            ##                                                self.c.winfo_reqheight())
        if event.width>=(scrollr[2]-scrollr[0]):
            # print "x_scrollbar away"
            self.sb_x.grid_remove()
        else:
            # print "x_scrollbar needed"
            self.sb_x.grid(row=1,column=0,sticky=Tkinter.EW)
            
        if event.height>=(scrollr[3]-scrollr[1]):
            # print "y_scrollbar away"
            self.sb_y.grid_remove()
        else:
            # print "y_scrollbar needed"
            self.sb_y.grid(row=0,column=1,sticky=Tkinter.NS)
            
            # hidden definitions for geometry methods
            
    def pack_configure(self, cnf={}, **kw):
        cnf=Tkinter._cnfmerge((cnf,kw))
        return self.hidden_frame.pack(cnf)
    pack = pack_configure
    
    def pack_forget(self):
        return self.hidden_frame.pack_forget()
    forget = pack_forget
    
    def pack_info(self):
        return self.hidden_frame.pack_info()
        
    def place_configure(self, cnf={}, **kw):
        cnf=Tkinter._cnfmerge((cnf,kw))
        return self.hidden_frame.place(cnf=cnf)
    place = place_configure
    
    def place_forget(self):
        return self.hidden_frame.place_forget()
    forget = place_forget
    
    def place_info(self):
        return self.hidden_frame.place_info()
        
    def grid_configure(self, cnf={}, **kw):
        cnf=Tkinter._cnfmerge((cnf,kw))
        print cnf
        return self.hidden_frame.grid(cnf=cnf)
    grid = grid_configure
    
    def grid_forget(self):
        return self.hidden_frame.grid_forget()
    forget = grid_forget
    
    def grid_info(self):
        return self.hidden_frame.grid_info()
        
    def grid_location(self, x, y):
        return self.hidden_frame.grid_location(x,y)
        
    location = grid_location
    
class tab_frame(Tkinter.Frame):
    """
    tabbed widgets
    """
    def tab_selected(self,event):
        """
        event handler for tabs
        """
        selected_item=event.widget.find_withtag(Tkinter.CURRENT)[0]
        selected_tags=event.widget.gettags(selected_item)
        tab_name=filter(lambda t: t[:4]=='tab_',selected_tags)[0][9:]
        key=ProbEditorBasics.tag_to_key(tab_name)
        if key!=self.actual_tab:
            self.change_tab(key)
            
    def change_tab(self,key):
        """
        process of changing:
        
        - lifts actual tab
        
        - changes color of previous tab to background
        
        - actual tab is marked white
        
        - releases (forget) previous widget
        
        - pack actual widget
        """
        # change tabs
        if self.actual_tab!=None:
            poly_list=self.tabs.find_withtag(\
                'tab_poly_'+ProbEditorBasics.key_to_tag(self.actual_tab))
            for item in poly_list:
                self.tabs.itemconfigure(item,fill='')
                
        tag=ProbEditorBasics.key_to_tag(key)
        poly_list=self.tabs.find_withtag('tab_poly_'+tag)
        for item in poly_list:
            self.tabs.itemconfigure(item,fill='white')
            self.tabs.lift(item)
        text_list=self.tabs.find_withtag('tab_text_'+tag)
        for item in text_list:
            self.tabs.lift(item)
            
            # change body
        if self.actual_tab!=None:
            self.widget_dict[self.actual_tab].pack_forget()
        self.widget_dict[key].pack(side=Tkinter.BOTTOM,
                                   in_=self.container,
                                   expand=1,fill=Tkinter.BOTH)
        self.actual_tab=key
        
    def __init__(self,master,widget_dict,start=None,**config):
        """
        init frame, containing:
        
        - tabbs
        
        - frame with fixed width, height for switched widgets
        
        size from slaves is not propagated
        """
        Tkinter.Frame.__init__(self,master,config)
        
        self.lower()
        self.tabs=Tkinter.Canvas(self,height=25,highlightthickness=0)
        # why are sizes here?
        self.container=Tkinter.Frame(self,width=400,height=300)
        self.container.pack_propagate(0)
        
        self.widget_dict=widget_dict
        key_list=widget_dict.keys()
        key_list.sort()
        text_base_y=20
        pos_x=7
        margin_x=3
        margin_y=3
        text_anchor=Tkinter.SW
        for key in key_list:
            # Reiter malen
            # erst den Text
            tab_text=key
            tag=ProbEditorBasics.key_to_tag(key)
            text=self.tabs.create_text((pos_x,text_base_y),
                                       text=key,
                                       anchor=text_anchor,
                                       tags=('tab_text_'+tag))
            text_box=self.tabs.bbox(text)
            
            #dann den Rand:
            far_right=2000
            far_left=-1
            very_deep=2000
            tab=self.tabs.create_polygon((far_left,text_base_y,
                                          text_box[0]-2-margin_x,text_base_y,
                                          text_box[0]-2,text_box[1]-margin_y,
                                          text_box[2],text_box[1]-margin_y,
                                          text_box[2]+margin_x,text_base_y,
                                          far_right,text_base_y,
                                          far_right,very_deep,
                                          far_left,very_deep),
                                         fill='',
                                         outline='black',
                                         tags=('tab_poly_'+tag)
                                         )
            # Text nach oben
            self.tabs.lift(text)
            pos_x=pos_x+text_box[2]-text_box[0]+2*margin_x
            
            # Event-Handler
            self.tabs.tag_bind(text,'<Button-1>',self.tab_selected)
            self.tabs.tag_bind(tab,'<Button-1>',self.tab_selected)
            
            # setze aktuellen Tab
        self.actual_tab=None
        if start!=None:
            self.change_tab(start)
        else:
            self.change_tab(key_list[0])
            
        self.tabs.pack(side=Tkinter.TOP,fill=Tkinter.X)
        self.container.pack(side=Tkinter.TOP,expand=1,fill=Tkinter.BOTH)
        
        #####################################################################################
        
class flyout_decoration:
    """
    works only with canvas derivatives
    """
    def __init__(self,info_function):
        """
        prepare bindings and store info-function
        """
        self.info_function=info_function
        # prepare bindings
        self.add_bindings(info_function)
        
    def add_bindings(self,info_function):
        """
        add bindings to canvas-items with 'flyout_info' tag
        """
        items=self.find_withtag(Tkinter.ALL)
        # init status
        self.flyout_stat=0
        # init id of after-timer
        self.after_id=0
        for item in items:
            key_tag=filter(lambda t:t=='flyout_info',self.gettags(item))
            if len(key_tag)==0: continue
            self.tag_bind(item,'<Enter>',self.flyout_enter,'+')
            self.tag_bind(item,'<Motion>',self.flyout_delay_start,'+')
            self.tag_bind(item,'<Leave>',self.flyout_leave,'+')
            
    def flyout_enter(self,event):
        """
        a item is entered, initialise status
        """
        # stat:
        # 0 no timer set, no flyout, after motion =>1
        # 1 timer set, no flyout, after motion =>1, no motion =>2
        # 2 flyout, no timer, after motion =>1
        if self.flyout_stat!=0:
            return
        self.flyout_stat=0
        self.after_id=0
        
    def flyout_leave(self,event):
        """
        the item is left:
        
        - cancel timer
        
        - hide flyout
        
        - reset status
        """
        if self.flyout_stat==1:
            self.after_cancel(self.after_id)
        if self.flyout_stat==2:
            self.flyout_hide()
        self.flyout_stat=0
        
    def flyout_delay_start(self,event):
        """
        mouse move over item:
        
        - store item
        
        - evtl. hide flyout and cancel previous timer
        
        - start new timer for move-timeout
        
        """
        item=self.find_withtag(Tkinter.CURRENT)[0]
        if self.flyout_stat==1:
            # moved again, no flyout set
            self.after_cancel(self.after_id)
        elif self.flyout_stat==2:
            # flyout set
            self.flyout_hide()
            # moved
        self.flyout_stat=1
        self.after_id=self.after(1000,self.flyout_delay_end,item,event.x,event.y)
        
    def flyout_delay_end(self,item,x,y):
        """
        after-event handler:
        
        no move occured, we can create a flyout
        """
        self.flyout_stat==2
        self.flyout_show(item,x,y)
        
    def flyout_show(self,item,x,y):
        """
        create the flyout, get info text from 'info_function'
        """
        self.flyout_stat=2
        text_x=self.canvasx(x)+15
        text_y=self.canvasy(y)+15
        flyout_text=self.info_function(item)
        text_item=self.create_text((text_x,text_y),text=flyout_text,tags=('flyout'),
                                   anchor=Tkinter.NW)
        self.create_rectangle(self.bbox(text_item),fill='khaki',tags=('flyout'))
        self.lift(text_item)
        
    def flyout_hide(self):
        """
        destroy flyout, reset state
        """
        self.flyout_stat=0
        flyout_elements=self.find_withtag('flyout')
        for element in flyout_elements:
            self.delete(element)
            
            ######################################################################################
            
class bar_chart_y(Tkinter.Canvas,flyout_decoration):
    """
    
    """
    def __init__(self,master,prob_dict,keys,colors,report_func=None):
        """
        
        """
        self.max_value=0
        self.max_key=None
        for k in keys:
            if prob_dict[k]>self.max_value:
                self.max_value=prob_dict[k]
                self.max_key=k
                
        self.bar_step=30
        self.bar_width=20
        self.text_length=30
        self.x_margin=10
        self.y_margin=10
        self.bar_length=350
        if self.max_value!=0:
            self.bar_factor=self.bar_length/self.max_value
        else:
            self.bar_factor=1.0
        Tkinter.Canvas.__init__(self,
                                master,
                                bg='white',
                                highlightthickness=0,
                                )
        self.create_description(prob_dict,keys)
        self.create_bars(prob_dict,keys,colors)
        self.report_func=None
        if report_func!=None:
            self.report_func=report_func
        self.config_bar_parameters(prob_dict)
        self.config_bar_height(prob_dict)
        flyout_decoration.__init__(self,self.info_func)
        self.bind('<Configure>',self.bars_configure)
        
    def info_func(self,item):
        tag=filter(lambda t:t[:4]=='tag_',self.gettags(item))[0]
        key=ProbEditorBasics.tag_to_key(tag[4:])
        bar=filter(lambda i,s=self:'bar' in s.gettags(i),self.find_withtag(tag))[0]
        coords=self.coords(bar)
        value=float(coords[2]-coords[0])/self.bar_factor
        if value>0:
            precision=-math.floor(math.log10(value))+3
        else:
            precision=1
        flyout_text=('%s: %s')%(key,str(round(value,precision)))
        return flyout_text
        
    def create_description(self,prob_dict,keys):
        text_anchor=Tkinter.W
        i=0
        self.text_length=0.0
        for k in keys:
            tag='tag_'+ProbEditorBasics.key_to_tag(k)
            pos_y=self.y_margin+i*self.bar_step
            text=self.create_text(self.x_margin,pos_y+self.bar_width/2,
                                  anchor=text_anchor,
                                  text=k,
                                  tags=('text',tag))
            text_box=self.bbox(text)
            this_length=abs(text_box[0]-text_box[2])
            if this_length>self.text_length:
                self.text_length=this_length
            i=i+1
            
    def create_bars(self,prob_dict,keys,colors):
        start_x=self.x_margin+self.text_length
        start_y=self.y_margin
        text_anchor=Tkinter.W
        i=0
        for k in keys:
            tag='tag_'+ProbEditorBasics.key_to_tag(k)
            pos_x=start_x
            pos_y=start_y+i*self.bar_step
            self.create_rectangle(pos_x,
                                  pos_y,
                                  pos_x,
                                  pos_y+self.bar_width,
                                  fill=colors[i],
                                  tags=('bar','flyout_info',tag))
            i=i+1
            
    def get_max_value(self):
        dict=self.get_bar_values()
        max_value=0.0
        max_key=''
        for k in dict.keys():
            if dict[k]>self.max_value:
                max_value=dict[k]
                max_key=k
        return (max_key,max_value)
        
    def get_bar_values(self,key_list=None):
        dict={}
        if not key_list:
            # return all keys
            bar_items=self.find_withtag('bar')
            for bar_item in bar_items:
                tags=self.gettags(bar_item)
                tag=filter(lambda t:'tag_'==t[:4],tags)[0][4:]
                key=ProbEditorBasics.tag_to_key(tag)
                coords=self.coords(bar_item)
                value=float(coords[2]-coords[0])/self.bar_factor
                dict.update({key:value})
        else:
            # return only keys in list
            for key in key_list:
                tag='tag_'+ProbEditorBasics.key_to_tag(key)
                items=self.find_withtag(tag)
                if len(items)==0: continue
                bar_item=filter(lambda i,s=self:'bar' in s.gettags(i),items)
                coords=self.coords(bar_item)
                value=float(coords[2]-coords[0])/self.bar_factor
                dict.update({key:value})
        return dict
        
    def bars_configure(self,event):
        actual_values=self.config_bar_parameters(width=event.width)
        self.update_bars(actual_values)
        
    def config_bar_parameters(self,prob_dict=None,width=None):
        # merge new values (if exist)
        actual_values=self.get_bar_values()
        if prob_dict!=None:
            for key in prob_dict.keys():
                if actual_values.has_key(key):
                    actual_values[key]=prob_dict[key]
        old_max=self.max_value
        # find new max value
        self.max_value=0.0
        self.max_key=''
        for key in actual_values.keys():
            if actual_values[key]>self.max_value:
                self.max_value=actual_values[key]
                self.max_key=key
                # expand bars (if necessary, fix to 100)
        if width!=None:
            self.bar_length=width-self.text_length-2*self.x_margin
        if self.bar_length<100:
            self.bar_length=100
            # decide to change scale
        if self.max_value*self.bar_factor>self.bar_length or \
           self.max_value*self.bar_factor<self.bar_length/2.0:
            # scale change, if it is rational
            if self.max_value!=0:
                self.bar_factor=self.bar_length/self.max_value
                if self.report_func!=None:
                    self.report_func('scale change',None,None)
        else:
            # no scale change
            actual_values=prob_dict
        return actual_values
        
    def config_bar_height(self,prob_dict):
        # do real work
        if prob_dict==None: return
        for key in prob_dict.keys():
            tag='tag_'+ProbEditorBasics.key_to_tag(key)
            items=self.find_withtag(tag)
            if len(items)==0: continue
            bar_item=filter(lambda i,s=self:'bar' in s.gettags(i),items)
            coords=self.coords(bar_item)
            self.coords(bar_item,coords[0],coords[1],
                        coords[0]+prob_dict[key]*self.bar_factor,coords[3])
            
    def config_bar_color(self,keys,color_list):
        bar_items=self.find_withtag('bar')
        for item in bar_items:
            tag=filter(lambda t:t[:4]=='tag_',self.gettags(item))[0][4:]
            key=ProbEditorBasics.tag_to_key(tag[4:])
            color=color_list[keys.index(key)]
            self.itemconfig(item,fill=color)
            
    def config_bar_order(self,keys):
        text_x=self.bar_width/2
        start_y=self.y_margin
        i=0
        for k in keys:
            tag='tag_'+ProbEditorBasics.key_to_tag(key)
            items=self.find_withtag(tag)
            if len(items)!=1:
                continue
            bar_item=filter(lambda i,s=self:s.type(i)=='rectangle',items)
            text_item=filter(lambda i,s=self:s.type(i)=='text',items)
            pos_y=start_y+i*self.bar_step
            bar_coords=self.coords(bar_item)
            self.coords(bar_item,
                        bar_coords[0],
                        pos_y,
                        bar_coords[2],
                        pos_y+self.bar_width)
            text_coords=self.coords(text_item)
            self.coords(text_item,text_coords[0],pos_y+text_y)
            i=i+1
            
            #######################################################################################
            
class e_bar_chart_y(bar_chart_y):

    def __init__(self,master,prob_dict,keys,colors,report_func):
        bar_chart_y.__init__(self,master,prob_dict,keys,colors)
        # report value changes to this function
        self.report_func=report_func
        self.normal_cursor=self.cget('cursor')
        self.current_item=None
        self.create_handles()
        flyout_decoration.add_bindings(self,self.info_func)
        self.config_handle_pos(prob_dict)
        
    def create_handles(self):
        for item in self.find_withtag('bar'):
            coords=self.coords(item)
            tags=self.gettags(item)
            tag=filter(lambda tag: tag[:4]=='tag_',tags)[0]
            handle=self.create_rectangle((coords[2],coords[1],coords[2],coords[3]),
                                         fill='black',
                                         tags=('bar_top','flyout_info',tag))
            self.tag_bind(handle,'<Button-1>',self.handle_mouse_move_start)
            self.tag_bind(handle,'<Enter>',self.change_cursor)
            self.tag_bind(handle,'<Leave>',self.restore_cursor)
            
    def info_func(self,item):
        tags=self.gettags(item)
        if 'bar_top' in tags:
            tag=filter(lambda t:t[:4]=='tag_',tags)[0]
            key=ProbEditorBasics.tag_to_key(tag[4:])
            coords=self.coords(item)
            value=self.value_from_handle_coords(coords)
            if value>0:
                precision=-math.floor(math.log10(value))+3
            else:
                precision=1
            flyout_text=('%s: %s')%(key,str(round(value,precision)))
        else:
            flyout_text=bar_chart_y.info_func(self,item)
        return flyout_text
        
    def change_cursor(self,event):
        self.on_handle=1
        event.widget.configure(cursor='sb_h_double_arrow')
        
    def restore_cursor(self,event):
        self.on_handle=0
        if self.current_item==None:
            event.widget.configure(cursor=self.normal_cursor)
            
    def value_from_handle_coords(self,coords):
        return float((coords[2]+coords[0])/2.0-self.x_margin-self.text_length)/self.bar_factor
        
    def handle_mouse_move_start(self,event):
        #count moves since region of maximum value
        self.move_above=1
        #count moves since maximum position
        self.move_beneath=1
        #cache information
        self.current_item=self.find_withtag(Tkinter.CURRENT)[0]
        tags=self.gettags(self.current_item)
        self.current_tag=filter(lambda tag: tag[:4]=='tag_',tags)[0]
        self.current_key=ProbEditorBasics.tag_to_key(self.current_tag[4:])
        
        values=self.get_bar_values()
        # for undo function
        self.report_func('old value',self.current_key,values[self.current_key])
        
        #find maximum value, except own value
        keys=values.keys()
        keys.remove(self.current_key)
        self.other_max=0.0
        for key in keys:
            if values[key]>self.other_max:
                self.other_max=values[key]
                
                #bind necessary motion-functions
        self.bind('<ButtonRelease-1>',self.handle_mouse_move_end)
        self.bind('<B1-Motion>',self.handle_mouse_move)
        
    def handle_mouse_move(self,event):
        x=self.canvasx(event.x)
        new_value=float(x-self.x_margin-self.text_length)/self.bar_factor
        upper_value=float(self.bar_length)/self.bar_factor
        if new_value<upper_value:
            self.move_beneath+=1
        if new_value>self.other_max:
            self.move_above+=1
            
        if new_value<=0:
            new_value=0
        elif new_value>upper_value:
            new_value=upper_value
            if self.move_beneath>0:
                actual_values=self.get_bar_values()
                self.bar_factor/=1.4142135623730951
                self.config_bar_height(actual_values)
                self.report_func('scale change',None,None) 
            self.move_beneath=0
        elif new_value<self.other_max*0.9 and \
             self.bar_factor!=self.bar_length/self.other_max:
            if self.move_above>0:
                actual_values=self.get_bar_values()
                self.bar_factor*=1.4142135623730951
                if self.bar_factor>self.bar_length/self.other_max:
                    self.bar_factor=self.bar_length/self.other_max
                self.config_bar_height(actual_values)
                self.report_func('scale change',None,None) 
            self.move_above=0
        self.config_handle_pos({self.current_key:new_value})
        self.report_func('move',self.current_key,new_value)
        
    def handle_mouse_move_end(self,event):
        self.unbind('<ButtonRelease-1>')
        self.unbind('<B1-Motion>')
        coords=self.coords(self.current_item)
        new_value=self.value_from_handle_coords(coords)
        self.config_bar_height({self.current_key:new_value})
        self.report_func('new value',self.current_key,new_value)
        self.current_item=None
        if self.on_handle==0:
            event.widget.configure(cursor=self.normal_cursor)
            
    def config_handle_pos(self,prob_dict):
        if prob_dict==None: return
        for key in prob_dict.keys():
            tag='tag_'+ProbEditorBasics.key_to_tag(key)
            items=self.find_withtag(tag)
            bar_item=filter(lambda i,s=self:'bar_top' in s.gettags(i),items)
            if len(bar_item)==0: continue
            coords=self.coords(bar_item)
            pos_x=self.x_margin+self.text_length+prob_dict[key]*self.bar_factor
            self.coords(bar_item,
                        pos_x-1,
                        coords[1],
                        pos_x+1,
                        coords[3])
            
    def config_bar_height(self,prob_dict):
        bar_chart_y.config_bar_height(self,prob_dict)
        self.config_handle_pos(prob_dict)
        
    def update_bars(self,prob_dict):
        actual_values=self.config_bar_parameters(prob_dict)
        self.config_bar_height(actual_values)
        
        ####################################################################################
        
class scale(Tkinter.Canvas):

    def __init__(self,master,start_x,factor,max_value=None):
        Tkinter.Canvas.__init__(self,
                                master,
                                height=25,
                                bg='white',
                                highlightthickness=0
                                )
        self.draw_arrow(start_x,factor,max_value)
        
    def draw_arrow(self,start_x,factor,max_value=None):
        self.create_line(start_x,
                         20,
                         start_x+factor*max_value+10,
                         20,
                         arrow=Tkinter.LAST,
                         tags=('arrow'))
        if max_value!=None:
            self.config_scale(start_x,factor,max_value)
            
    def config_scale(self,start_x,factor,max_value):
        tic_tags=self.find_withtag('tic')
        for tag in tic_tags:
            self.delete(tag)
        arrow=self.find_withtag('arrow')
        self.coords(arrow,
                    start_x,
                    20,
                    start_x+factor*max_value+10,
                    20)
        scale_frac=10.0**(math.floor(math.log10(max_value)))
        if scale_frac*factor>100:
            scale_frac/=2
        if scale_frac*factor<50:
            scale_frac*=2
            # XXX: the scaling needs some fine tuning
        round_pos=int(-math.floor(math.log10(scale_frac)))
        x_pos=0.0
        while x_pos<=max_value:
            pixel_x=start_x+x_pos*factor
            self.create_line(pixel_x,15,pixel_x,25,
                                   tags=('tic'))
            self.create_text(pixel_x,15,text=str(round(x_pos,round_pos)),
                             anchor=Tkinter.S,
                             tags=('tic','text'))
            x_pos=x_pos+scale_frac
            
            
            ####################################################################################
            
class bar_chart_with_scale(Tkinter.Frame):

    def __init__(self,master,prob_dict,keys,colors,report_func):
        """
        glue scale and barchart together
        """
        Tkinter.Frame.__init__(self,master)
        self.bars=e_bar_chart_y(self,prob_dict,keys,colors,self.bar_report)
        scale_start_x=self.bars.x_margin+self.bars.text_length
        factor=self.bars.bar_factor
        scale_max=self.bars.max_value
        self.scale=scale(self,scale_start_x,factor,scale_max)
        
        self.report_func=report_func
        self.scrollbar = Tkinter.Scrollbar(self,orient=Tkinter.VERTICAL,
                                           command=self.bars.yview)
        region=self.bars.bbox(Tkinter.ALL)
        sregion=(0,0,region[2],region[3])
        self.bars.config(scrollregion=sregion,
                         yscrollcommand=self.scrollbar.set)
        
        empty=Tkinter.Frame(self,bg='white')
        self.scale.grid(column=0,row=0,sticky=Tkinter.E+Tkinter.W)
        self.bars.grid(column=0,row=1,sticky=Tkinter.E+Tkinter.W+Tkinter.N+Tkinter.S)
        empty.grid(column=1,row=0,sticky=Tkinter.E+Tkinter.W+Tkinter.N+Tkinter.S)
        Tkinter.Frame.columnconfigure(self,0,weight=1)
        Tkinter.Frame.rowconfigure(self,1,weight=1)
        self.bind('<Configure>',self.config_event)
        
    def config_event(self,event):
        """
        add or remove scrollbar
        """
        scrollr=self.bars.cget('scrollregion')
        scrollr=map(int,scrollr.split())
        # print "from config",scrollr
        #hier muesste noch ein config_scale rein
        # self.scale.config_scale()
        if event.height>=(scrollr[3]-scrollr[1]):
            # print "y_scrollbar away"
            self.scrollbar.grid_remove()
        else:
            # print "y_scrollbar needed"
            self.scrollbar.grid(column=1,row=1,sticky=Tkinter.NS)
            
    def bar_report(self,what,key,value):
        if what=='move':
            return
        if what=='scale change':
            self.scale.config_scale(self.bars.x_margin+self.bars.text_length,
                                    self.bars.bar_factor,
                                    self.bars.bar_length/self.bars.bar_factor)
            return
            
        self.report_func(what,key,value)
        
        
        ####################################################################################
        
class rod_chart(Tkinter.Canvas):
    def __init__(self,master,prob_dict,keys,colors):
        self.ProbDict=prob_dict
        self.rod_box=(10,10,300,60)
        Tkinter.Canvas.__init__(self,master,
                                bg='white',highlightthickness=0,
                                width=200,height=320
                                )
        self.draw_rods(keys,colors)
        
    def draw_rods(self,keys,colors):
        start_pos=self.rod_box[0]
        factor=(self.rod_box[2]-self.rod_box[0])/self.ProbDict.sum
        i=0
        for key in keys:
            end_pos=start_pos+factor*self.ProbDict[key]
            self.create_rectangle(start_pos,
                                  self.rod_box[1],
                                  end_pos,
                                  self.rod_box[3],
                                  fill=colors[i],
                                  tags=('rod','tag_'+key))
            i=i+1
            
        for key in keys:
            rod_item=self.find_withtag('tag_'+key)[0]
            coords=self.coords(rod_item)
            text_line_x=(coords[0]+coords[2])/2
            h1=0.7
            h2=1.2
            h3=1.5
            text_line_y1=(coords[1]*(1-h1)+coords[3]*h1)
            text_line_y2=(coords[1]*(1-h2)+coords[3]*h2)
            text_y=(coords[1]*(1-h3)+coords[3]*h3)
            self.create_line(text_line_x,text_line_y1,
                             text_line_x,text_line_y2,
                             tags=('tag_'+key,'line'))
            self.create_text(text_line_x,text_y,
                             text=key,
                             anchor=Tkinter.N)
            
class e_rod_chart(rod_chart):
    def __init__(self,master,prob_dict,keys,colors):
        rod_chart.__init__(self,master,prob_dict,keys,colors)
        
        self.text_item=self.create_text(20,190,text='Los gehts',tag='info')
        
        for item in self.find_withtag('bar'):
            self.tag_bind(item,'<Button-1>',self.handle_mouse_move_start)
            
    def handle_mouse_move_start(self,event):
        self.current_item=self.find_withtag(Tkinter.CURRENT)[0]
        coords=self.coords(self.current_item)
        if self.canvasy(event.y)>coords[1]+2:
            return
        self.bind('<ButtonRelease-1>',self.handle_mouse_move_end)
        self.bind('<B1-Motion>',self.handle_mouse_move)
        self.itemconfigure(self.text_item, text='Enter')
        
    def handle_mouse_move(self,event):
        coords=self.coords(self.current_item)
        y=self.eventy(event.y)
        if y>coords[3] or y<self.bar_box[1]:
            return
        coords[1]=y
        self.coords(self.current_item,tuple(coords))
        self.itemconfigure(self.text_item, text='Move')
        
    def handle_mouse_move_end(self,event):
        self.unbind('<ButtonRelease-1>')
        self.unbind('<B1-Motion>')
        self.itemconfigure(self.text_item, text='Leave')
        tags=self.gettags(self.current_item)
        tag=filter(lambda tag: tag[:4]=='tag_',tags)[0]
        coords=self.coords(self.current_item)
        new_value=float(coords[1]-coords[3])/self.bar_height
        self.change_value(tag[4:],new_value)
        
    def change_value(self,key,value):
        self.ProbDict[key]=value
        self.ProbDict.changed()
        
        ####################################################################################
        
class pie_chart(Tkinter.Canvas,flyout_decoration):

    def __init__(self,master,prob_dict,keys,colors):
        self.ProbDict=prob_dict
        self.arc_box=(100,50,300,250)
        self.key_list=keys
        Tkinter.Canvas.__init__(self,master,bg='white',highlightthickness=0)
        self.draw_arcs(self.arc_box,keys,colors)
        flyout_decoration.__init__(self,self.info_function)
        
    def info_function(self,item):
        tag=filter(lambda t:t=='other' or t[:4]=='tag_',self.gettags(item))[0]
        sector=filter(lambda i,s=self:'sector' in s.gettags(i),self.find_withtag(tag))
        if tag!='other':
            key=ProbEditorBasics.tag_to_key(tag[4:])
        else:
            key='other'
        value=abs(float(self.itemcget(sector,'extent')))/360.0*self.ProbDict.sum
        l10_value=-math.floor(math.log10(value))
        info='%s: %s'%(key,str(round(value,int(l10_value+3))))
        return info
        
    def anchor(self,angle,angle_base):
        """
        returns anchor-position for text
        angle is counted anti-clockwise from 3o'clock
        """
        # Winkel auf jeden Fall positiv zaehlen
        if angle<0:
            my_angle=angle_base/2-angle
        my_angle=(angle/angle_base-math.floor(angle/angle_base))*8.0
        if my_angle<1 or my_angle>=7:
            return Tkinter.W
        if my_angle<2:
            return Tkinter.SW
        if my_angle<3:
            return Tkinter.SE
        if my_angle<5:
            return Tkinter.E
        if my_angle<6:
            return Tkinter.NE
        return Tkinter.NW
        
    def draw_arcs(self,circle_box,key_list,colors):
        """
        """
        this_sum=0.0
        #Start bei 6 Uhr
        start_angle=270.0
        # im Uhrzeigersinn laufen
        full_angle=-360.0
        i=0
        while i<=len(key_list):
            key=''
            tag=''
            this_angle=0
            extent_angle=0
            if i<len(key_list):
                key=key_list[i]
                tag='tag_'+ProbEditorBasics.key_to_tag(key)
                this_angle=start_angle+this_sum*full_angle/self.ProbDict.sum
                extent_angle=self.ProbDict[key]*full_angle/self.ProbDict.sum
                this_sum=this_sum+self.ProbDict[key]
                color=colors[i]
            else:
                key='other'
                tag='other'
                if 1.0-this_sum/self.ProbDict.sum<=1e-5:
                    break
                this_angle=start_angle+this_sum*full_angle/self.ProbDict.sum
                extent_angle=start_angle+full_angle-this_angle
                color=self.cget('bg')
            line_angle=0.0
            if abs(extent_angle)>=360:
                # fast voller Kreis
                extent_angle=359.9
                # Volle Kreise zeichnen
            sector=self.create_arc(circle_box,
                                   start=this_angle,
                                   extent=extent_angle,
                                   tags=('sector','flyout_info',tag),
                                   fill=color)
            i=i+1
            
        for item in self.find_withtag('sector'):
            coords=self.coords(item)
            start_angle=float(self.itemcget(item,'start'))
            extent_angle=float(self.itemcget(item,'extent'))
            line_angle=start_angle+extent_angle/2
            tags=filter(lambda t: t[:4]=='tag_',self.gettags(item))
            text=''
            if len(tags)==0:
                text='other'
                tags=('other',)
            else:
                text=ProbEditorBasics.tag_to_key(tags[0][4:])
            m_x=math.cos(line_angle/180.0*math.pi)
            m_y=-math.sin(line_angle/180.0*math.pi)
            circle_middle_x=(coords[0]+coords[2])/2
            circle_middle_y=(coords[1]+coords[3])/2
            circle_diam_x=(coords[2]-coords[0])/2
            circle_diam_y=(coords[3]-coords[1])/2
            self.create_line(circle_middle_x+m_x*circle_diam_x*0.8,
                             circle_middle_y+m_y*circle_diam_y*0.8,
                             circle_middle_x+m_x*circle_diam_x*1.1,
                             circle_middle_y+m_y*circle_diam_y*1.1,
                             tags=('line',tags))
            # hier ist der Text!
            self.create_text(circle_middle_x+m_x*circle_diam_x*1.1,
                             circle_middle_y+m_y*circle_diam_y*1.1,
                             anchor=self.anchor(line_angle,360.0),
                             text=text,
                             tags=('text',tags))
            
    def configure_description(self,item):
        coords=self.coords(item)
        if len(coords)==0: return
        start_angle=float(self.itemcget(item,'start'))
        extent_angle=float(self.itemcget(item,'extent'))
        line_angle=start_angle+extent_angle/2
        m_x=math.cos(line_angle/180.0*math.pi)
        m_y=-math.sin(line_angle/180.0*math.pi)
        circle_middle_x=(coords[0]+coords[2])/2
        circle_middle_y=(coords[1]+coords[3])/2
        circle_diam_x=(coords[2]-coords[0])/2
        circle_diam_y=(coords[3]-coords[1])/2
        tags=filter(lambda t: t[:4]=='tag_',self.gettags(item))
        if len(tags)==0:
            tags=('other',)
            
        tag_items=self.find_withtag(tags[0])
        line_item=text_item=0
        for tag_item in tag_items:
            item_type=self.type(tag_item)
            if item_type=="line":
                line_item=tag_item
            elif item_type=="text":
                text_item=tag_item
                
        self.coords(line_item,
                    circle_middle_x+m_x*circle_diam_x*0.8,
                    circle_middle_y+m_y*circle_diam_y*0.8,
                    circle_middle_x+m_x*circle_diam_x*1.1,
                    circle_middle_y+m_y*circle_diam_y*1.1)
        # hier ist der Text!
        self.coords(text_item,
                    circle_middle_x+m_x*circle_diam_x*1.1,
                    circle_middle_y+m_y*circle_diam_y*1.1)
        self.itemconfig(text_item,
                        anchor=self.anchor(line_angle,360.0))
        
    def update_colors(self,order_list,color_list):
        sectors=self.find_withtag('sector')
        for sector in sectors:
            tag=filter(lambda t:t[:4]=='tag_',self.gettags(sector))[0]
            key=ProbEditorBasics.tag_to_key(tag[4:])
            if len(keytag)!=1: continue
            color=color_list[order_list.index(key)]
            self.itemconfig(sector,color=color)
            
    def update_position(self,dict,order_list):
        this_sum=0.0
        #Start bei 6 Uhr
        start_angle=270.0
        # im Uhrzeigersinn laufen
        full_angle=-360.0
        i=0
        while i<=len(order_list):
            key=''
            tag=''
            this_angle=0
            extent_angle=0
            if i<len(order_list):
                key=order_list[i]
                tag='tag_'+ProbEditorBasics.key_to_tag(key)
                this_angle=start_angle+this_sum*full_angle/dict.sum
                extent_angle=dict[key]*full_angle/dict.sum
                this_sum=this_sum+dict[key]
            else:
                key='other'
                tag='other'
                if 1.0-this_sum/dict.sum<=1e-5:
                    break
                this_angle=start_angle+this_sum*full_angle/dict.sum
                extent_angle=start_angle+full_angle-this_angle
            line_angle=0.0
            sector=filter(lambda item,s=self:'sector' in s.gettags(item),
                          self.find_withtag(tag))
            
            if abs(extent_angle)>=360:
                extent_angle=359.9
            self.itemconfig(sector,
                            start=this_angle,
                            extent=extent_angle)
            self.configure_description(sector)
            i=i+1
            
    def angle(self,x,y):
        if y==0:
            if x<0: return 180.0
            else: return 0.0
        elif x==0:
            if y<0: return 90.0
            else: return 270.0
        elif y<0:
            return math.atan(x/y)/math.pi*180.0+90.0
        else:
            return math.atan(x/y)/math.pi*180.0+270.0
            
            ######################################################################################
            
class e_pie_chart(pie_chart):

    def __init__(self,master,prob_dict,keys,colors,report_func):
        pie_chart.__init__(self,master,prob_dict,keys,colors)
        self.report_func=report_func
        self.oval_diam_x=self.oval_diam_y=4
        self.init_handles()
        flyout_decoration.add_bindings(self,self.info_function)
        
    def info_function(self,item):
        tags=self.gettags(item)
        if 'handle' in tags:
            tag=filter(lambda t:t[:4]=='tag_' or t=='other',tags)[0]
            if tag!='other':
                key=ProbEditorBasics.tag_to_key(tag[4:])
            else:
                key='other'
            arc=filter(lambda i,s=self:s.type(i)=='arc',
                        self.find_withtag(tag))
            neighbours=self.find_neighbours(arc)
            neighbours.append(arc)
            flyout_text=''
            for item in neighbours:
                flyout_text+=', '+pie_chart.info_function(self,item)
            flyout_text=flyout_text[2:]
        else:
            flyout_text=pie_chart.info_function(self,item)
        return flyout_text
        
        
    def find_neighbours(self,sector_item):
        # find neighbour sector
        angle_pos=float(self.itemcget(sector_item,'start'))
        sectors=self.find_withtag('sector')
        neighbours=[]
        for sector in sectors:
            extent_angle=float(self.itemcget(sector,'extent'))
            start_angle=float(self.itemcget(sector,'start'))
            angle_diff=divmod(start_angle+extent_angle-angle_pos,360.0)[1]
            if angle_diff<0.1 or angle_diff>359.9:
                neighbours.append(sector)
        return neighbours
        
    def move_start(self,event):
        # find sector belonging to handle
        self.current_item=self.find_withtag(Tkinter.CURRENT)[0]
        self.current_tag=filter(lambda t:t[:4]=='tag_' or t=='other',
                                self.gettags(self.current_item))[0]
        self.current_arc1=filter(lambda i,s=self:s.type(i)=='arc',
                                 self.find_withtag(self.current_tag))
        # find neighbour sector
        self.current_arc2=self.find_neighbours(self.current_arc1)
        
        # really found one?
        if len(self.current_arc2)==0:
            print 'no neighbour found'
            return
        if len(self.current_arc2)>1:
            print 'too much neighbours found'
            return
            
        self.current_arc2=self.current_arc2[0]
        # cache coordinates
        coords=self.coords(self.current_arc1)
        self.current_center_x=(coords[0]+coords[2])/2.0 #center of arc
        self.current_center_y=(coords[1]+coords[3])/2.0 #center of arc
        self.current_diam_x=abs(coords[1]-coords[3])/2.0
        self.current_diam_y=abs(coords[0]-coords[2])/2.0
        # save angle sum
        self.angle_start=float(self.itemcget(self.current_arc2,'start'))
        self.angle_extent=float(self.itemcget(self.current_arc1,'extent'))+float(self.itemcget(self.current_arc2,'extent'))
        
        self.bind('<B1-Motion>',self.move)
        self.bind('<ButtonRelease-1>',self.move_end)
        
    def move(self,event):
        # compute new angle
        canvas_x=self.canvasx(event.x)
        canvas_y=self.canvasy(event.y)
        angle=self.angle(canvas_x-self.current_center_x,canvas_y-self.current_center_y)
        # is this angle valid?
        # and calcualte new start,extent angles
        new_extent2=float(self.itemcget(self.current_arc2,'extent'))
        new_start1=float(self.itemcget(self.current_arc1,'start'))
        if self.angle_extent>0:
            #anti-clockwise orientation
            if self.angle_start>angle:
                if self.angle_start+self.angle_extent<=angle+360:
                    # valid
                    new_start1=angle
                    new_extent2=angle+360-self.angle_start
                else:
                    #invalid
                    pass
            else:
                if self.angle_start+self.angle_extent<=angle:
                    # valid
                    new_start1=angle
                    new_extent2=angle-self.angle_start
                else:
                    # invalid
                    pass
        else:
            #clockwise orientation
            if self.angle_start<angle:
                if self.angle_start+self.angle_extent<=angle-360:
                    # valid
                    new_start1=angle
                    new_extent2=angle-360-self.angle_start
                else:
                    # invalid
                    pass
            else:
                if self.angle_start+self.angle_extent<=angle:
                    # valid
                    new_start1=angle
                    new_extent2=angle-self.angle_start
                else:
                    # invalid
                    pass
                    
                    # set new arc angles
        new_start2=self.angle_start
        new_extent1=self.angle_extent-new_extent2
        # pitfall! extent>=360
        if new_extent1>359.9:
            new_extent1=359.9
        elif new_extent1<-359.9:
            new_extent1=-359.9
        if new_extent2>359.9:
            new_extent2=359.9
        elif new_extent2<-359.9:
            new_extent2=-359.9
        self.itemconfig(self.current_arc1,start=new_start1)
        self.itemconfig(self.current_arc2,start=new_start2)
        self.itemconfig(self.current_arc1,extent=new_extent1)
        self.itemconfig(self.current_arc2,extent=new_extent2)
        # set handle
        oval_x=math.cos(new_start1/180.0*math.pi)*self.current_diam_x+self.current_center_x
        oval_y=-math.sin(new_start1/180.0*math.pi)*self.current_diam_y+self.current_center_y
        oval=self.coords(self.current_item,
                         oval_x-self.oval_diam_x,oval_y-self.oval_diam_y,
                         oval_x+self.oval_diam_x,oval_y+self.oval_diam_y)
        
        # set description
        self.configure_description(self.current_arc1)
        self.configure_description(self.current_arc2)
        
        # report move
        tags=self.gettags(self.current_arc1)
        tag1=filter(lambda tag: tag[:4]=='tag_' or tag=='other',tags)[0]
        if tag1!='other':
            key1=ProbEditorBasics.tag_to_key(tag1[4:])
        else:
            key1='other'
        tags=self.gettags(self.current_arc2)
        tag2=filter(lambda tag: tag[:4]=='tag_' or tag=='other',tags)[0]
        if tag2!='other':
            key2=ProbEditorBasics.tag_to_key(tag2[4:])
        else:
            key2='other'
        value1=abs(float(self.itemcget(self.current_arc1,
                                       'extent')))/360.0*self.ProbDict.sum
        value2=abs(float(self.itemcget(self.current_arc2,
                                       'extent')))/360.0*self.ProbDict.sum
        report_dict={}
        report_dict[key1]=value1
        report_dict[key2]=value2
        self.report_func('move value',report_dict)
        
        
    def move_end(self,event):
        # no events!
        self.unbind('<B1-Motion>')
        self.unbind('<ButtonRelease-1>')
        # read new values from graphic
        tags=self.gettags(self.current_arc1)
        tag1=filter(lambda tag: tag[:4]=='tag_' or tag=='other',tags)[0]
        if tag1!='other':
            key1=ProbEditorBasics.tag_to_key(tag1[4:])
        else:
            key1='other'
        tags=self.gettags(self.current_arc2)
        tag2=filter(lambda tag: tag[:4]=='tag_' or tag=='other',tags)[0]
        if tag2!='other':
            key2=ProbEditorBasics.tag_to_key(tag2[4:])
        else:
            key2='other'
        value1=abs(float(self.itemcget(self.current_arc1,
                                       'extent')))/360.0*self.ProbDict.sum
        value2=abs(float(self.itemcget(self.current_arc2,
                                       'extent')))/360.0*self.ProbDict.sum
        report_dict={}
        report_dict[key1]=value1
        report_dict[key2]=value2
        self.report_func('new value',report_dict)
        self.current_tag=None
        self.current_key=None
        
        
    def init_handles(self):
        sectors=list(self.find_withtag('sector'))
        sectors.reverse()
        for sector in sectors:
            coords=self.coords(sector)
            tag=filter(lambda t:t[:4]=='tag_' or t=='other',self.gettags(sector))
            angle=float(self.itemcget(sector,'start'))
            center_x=(coords[0]+coords[2])/2.0 #center of arc
            center_y=(coords[1]+coords[3])/2.0 #center of arc
            diam_x=abs(coords[1]-coords[3])/2.0
            diam_y=abs(coords[0]-coords[2])/2.0
            oval_x=math.cos(angle/180.0*math.pi)*diam_x+center_x
            oval_y=-math.sin(angle/180.0*math.pi)*diam_y+center_y
            oval=self.create_oval(oval_x-self.oval_diam_x,oval_y-self.oval_diam_y,
                                  oval_x+self.oval_diam_x,oval_y+self.oval_diam_y,
                                  fill='black',tags=('handle','flyout_info',tag))
            self.tag_bind(oval,'<Button-1>',self.move_start)
            
    def update_handles(self):
        sectors=list(self.find_withtag('sector'))
        sectors.reverse()
        for sector in sectors:
            coords=self.coords(sector)
            tag=filter(lambda t:t[:4]=='tag_' or t=='other',self.gettags(sector))
            angle=float(self.itemcget(sector,'start'))
            center_x=(coords[0]+coords[2])/2.0 #center of arc
            center_y=(coords[1]+coords[3])/2.0 #center of arc
            diam_x=abs(coords[1]-coords[3])/2.0
            diam_y=abs(coords[0]-coords[2])/2.0
            oval_x=math.cos(angle/180.0*math.pi)*diam_x+center_x
            oval_y=-math.sin(angle/180.0*math.pi)*diam_y+center_y
            oval=filter(lambda i,s=self:s.type(i)=='oval',
                        self.find_withtag(tag))
            
            self.coords(oval,oval_x-self.oval_diam_x,oval_y-self.oval_diam_y,
                        oval_x+self.oval_diam_x,oval_y+self.oval_diam_y)
            
    def update_position(self,dict,order_list):
        pie_chart.update_position(self,dict,order_list)
        self.update_handles()



#!/usr/bin/env python2.6
################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#	file:   GatoFile.py
#	author: Achim Gaedke (achim.gaedke@zpr.uni-koeln.de)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################
import sys
import os
import types
import codecs
import StringIO
import urllib2
import xml.dom.minidom
import Gato
import Graph
import GraphUtil
import Tkinter
import TreeWidget
import TextTreeWidget
import ScrolledText

# code handling:
# the data object model module is unicode
# xml*Element functions return iso-8859-1 strings
# tkinter does not like python unicode, it gets iso-8859-1 strings
# file data are encoded to iso-8859-1

# xmlDecode(unicode) returns iso char string
(_isoEncoder,_isoDecoder,_isoStreamReader,_isoStreamWriter)=codecs.lookup("iso-8859-1")
xmlDecode=lambda x:_isoEncoder(x)[0]
# xmlEnocde(iso_char_string) returns unicode
xmlEncode=lambda x:_isoDecoder(x)[0]
# object that encodes stream data from unicode to iso-8859-1
xmlEncodedStream=_isoStreamWriter

class FileException(Exception):
    """
    is the subclass for all file and xml related things that can go wrong in this module
    """
    def __init__(self, reason):
        """
        takes the reason or explanation as argument
        """
        self.reason=reason
        
class GatoStorageCache:
    """
    singleton list for all accessed/used/cached storage locations
    """
    pass
    
class GatoStorageLocation:
    """
    bundle of common functions for file load/save/selection
    These functions act on the level of graphs and algorithms.
    they propagate the open and save calls to the different file implementations
    """
    
    def __init__(self,fileName=None):
        """
        opens the file and prepares the access
        """
        self.fileName=fileName
        
    def chooseLocation(self,accessMode="rw"):
        """
        opens a suitable dialog to choose a possible location for reading/writing gato data
        """
        
    def __repr__(self):
        """
        a short representation of the instance
        """
        return "<%s: %s>"%(self.__class__.__name__,self.fileName)
        
    def __str__(self):
        """
        should display the name in human understandable manner
        """
        return self.fileName
        
    def readGraph(self,whichGraph=None):
        """
        returns a Graph object
        """
        pass
        
    def readAlgortihm(self, whichAlgorithm=None):
        """
        returns an Algorithm object
        """
        pass
        
    def writeGraph(self, Graph, whichGraph=None):
        """
        puts a graph to the collection of graphs
        """
        pass
        
    def writeAlgorithm(self, Algorithm, whichAlgorithm=None):
        """
        puts a graph to the collection of algorithms
        """
        pass
        
class GatoStorageLocationTest:
    """
    some tests for GatoStorageLocation
    """
    def __init__(self):
        l=GatoStorageLocation("bla")
        print "human readable: %s, python representation %s"%(l,repr(l))
        
        
class GatoDirBranch(TreeWidget.DirBranch):
    """
    """
    def updateChildList(self):
        """
        create/update child list suitable for gato
        """
        oldChildren=self.children
        self.children=[]
        entries=[]
        try:
            entries=os.listdir(self.path)
        except OSError:
            pass
        entries.sort()
        oldEntries=map(lambda c:c.name,oldChildren)
        for entry in entries:
            if entry in oldEntries:
                self.children.append(oldChildren[oldEntries.index(entry)])
            else:
                entrypath=os.path.join(self.path,entry)
                if os.path.isdir(entrypath):
                    self.children.append(GatoDirBranch(parent=self,name=entry))
                elif os.path.isfile(entrypath):
                    # select suitable leaf
                    if entry[-5:]==".gato":
                        self.children.append(GatoFileBranch(parent=self,name=entry))
                    elif entry[-4:]==".xml":
                        self.children.append(xmlGatoFileBranch(parent=self,name=entry))
                    elif entry[-4:]==".alg" and os.path.isfile(entrypath[:-4]+".pro"):
                        self.children.append(algLeaf(parent=self,name=entry))
                    elif entry[-4:]==".cat":
                        self.children.append(catLeaf(parent=self,name=entry))
                        
class GatoFileBranch(TreeWidget.xmlFileBranch):
    """
    ToDo: supports xml files with gato root section
    """
    pass
    
class xmlGatoFileBranch(TreeWidget.xmlFileBranch):
    """
    ToDo supports xml files with gato sections inside
    """
    pass
    
class algLeaf(TreeWidget.Leaf):
    """
    old algortithm file
    """
    def __init__(self,parent=None,name=None):
        TreeWidget.Leaf.__init__(self,parent=parent, name=name)
        self.selectable=1
        
class catLeaf(TreeWidget.Leaf):
    """
    old graph file
    """
    def __init__(self,parent=None,name=None):
        TreeWidget.Leaf.__init__(self,parent=parent, name=name)
        self.selectable=1
        
class GatoTree(TreeWidget.scrolledTree):
    """
    """
    def __init__(self,master):
        TreeWidget.scrolledTree.__init__(self,master)
        self.root=GatoDirBranch(parent=self,name="/",anchor=(1,1))
        self.root.display()
        # show current work directory
        current=os.getcwd()
        currentList=[]
        while current!=self.root.name:
            (current,piece)=os.path.split(current)
            currentList.insert(0,piece)
            # expand all branches on the way
        thisNode=self.root
        thisNode.expand()
        for piece in currentList:
            thisNode=filter(lambda x:x.name==piece,thisNode.children)[0]
            thisNode.expand()
            
        (Nx0,Ny0,Nx1,Ny1)=thisNode.getBoundingBox()
        (Cx0,Cy0,Cx1,Cy1)=self.bbox(Tkinter.ALL)
        self.yview(Tkinter.MOVETO,float(Ny0)/float(Cy1-Cy0))
        
        
class WorkInProgress(Tkinter.Tk):
    """
    my development and test class
    """
    
    def __init__(self):
        Tkinter.Tk.__init__(self)
        self.tree=GatoTree(self)
        self.tree.pack(expand=1,fill=Tkinter.BOTH)
        
    def run(self):
        self.mainloop()
        
        
class xmlAlgorithmElement:

    algorithmElementName="algorithm"
    sourceElementName="text"
    prologElementName="prolog"
    aboutElementName="about"
    
    def __init__(self,dom):
        # test for right tag
        if dom.tagName!=self.algorithmElementName:
            raise FileException("Wrong element name: %s, %s expected"
                                %(dom.tagName,self.algorithmElementName))
        self.domElement=dom
        
    def setAlgorithmFromALGFile(self,fileName):
        """
        writes algorithm data to xml sections
        """
        algorithm=Gato.Algorithm()
        algorithm.Open(fileName)
        self.setAlgorithm(algorithm.source)
        self.setAbout(algorithm.About())
        PrologFileName=fileName[:-4]+'.pro'
        if os.access(PrologFileName,os.R_OK):
            self.setProlog(file(PrologFileName).read())
        else:
            print "Warning: could not read prolog file %s"%PrologFileName
            
    def setProlog(self,text):
        """
        creates/repaces the prolog section
        """
        #create new one
        newProlog=None
        if self.domElement.ownerDocument:
            newProlog=self.domElement.ownerDocument.createElement(self.prologElementName)
            newProlog.appendChild(self.domElement.ownerDocument.createCDATASection(
                xmlEncode(text)
                ))
        else:
            newProlog=xml.dom.Element(self.prologElementName)
            newProlog.appendChild(xml.dom.CDATASection(xmlEncode(text)))
        existingProlog=self.domElement.getElementsByTagName(self.prologElementName)
        if len(existingProlog)>0:
            # replace existing prolog
            self.domElement.repaceChild(newProlog,existingProlog[0])
        else:
            self.domElement.appendChild(newProlog)
            
    def setAlgorithm(self,text):
        """
        creates/repaces the algorithm section
        """
        #create new one
        newAlgorithm=None
        if self.domElement.ownerDocument:
            newAlgorithm=self.domElement.ownerDocument.createElement(self.sourceElementName)
            newAlgorithm.appendChild(self.domElement.ownerDocument.createCDATASection(
                xmlEncode(text)
                ))
        else:
            newAlgorithm=xml.dom.Element(self.sourceElementName)
            newAlgorithm.appendChild(xml.dom.CDATASection(xmlEncode(text)))
        existingAlgorithm=self.domElement.getElementsByTagName(self.sourceElementName)
        if len(existingAlgorithm)>0:
            # replace existing algorithm
            self.domElement.repaceChild(newAlgorithm,existingAlgorithm[0])
        else:
            self.domElement.appendChild(newAlgorithm)
            
    def setAbout(self,text):
        """
        ToDo
        """
        pass
        
    def getTextFromElement(self,tagName):
        "returns the text of specified tag"
        requestedElements=self.domElement.getElementsByTagName(tagName)
        if len(requestedElements)!=1:
            return None
        requestedElement=requestedElements[0]
        text=""
        for e in requestedElement.childNodes:
            if (e.nodeType==xml.dom.minidom.Node.TEXT_NODE or
                e.nodeType==xml.dom.minidom.Node.CDATA_SECTION_NODE):
                text+=e.data
        return xmlDecode(text)
        
    def getText(self):
        t=self.getTextFromElement(self.sourceElementName)
        if t:
            return t
        else:
            return ""
            
    def getAboutAsString(self):
        requestedElements=filter(lambda e:e.nodeType==xml.dom.Node.ELEMENT_NODE and \
                                 e.tagName==self.aboutElementName,
                                 self.domElement.childNodes)
        if len(requestedElements)<1:
            return None
        else:
            return requestedElements[0].toxml()
            
    def getTextAsFile(self):
        """
        creates a StringIO Object from file
        """
        return StringIO.StringIO(self.getText())
        
    def getProlog(self):
        "returns the prolog"
        p=self.getTextFromElement(self.prologElementName)
        if p:
            return p
        else:
            return ""
            
    def getPrologAsFile(self):
        return StringIO.StringIO(self.getProlog())
        
    def getName(self):
        return xmlDecode(self.domElement.getAttribute("name"))
        
    def setName(self,name):
        self.domElement.setAttribute("name",xmlEncode(name))
        
class xmlGraphElement:

    graphElementName="graph"
    
    def __init__(self,graph,name=None):
        """
        called with a dom element: grants access to Graph
        called with a graph and a name string: creates the dom structure from graph
        """
        # test for right tag
        if issubclass(graph.__class__,xml.dom.minidom.Element):
            # we were called with a dom element as argument
            if graph.tagName!=self.graphElementName:
                raise FileException("Wrong element name: %s, %s expected"
                                    %(graph.tagName,self.graphElementName))
            self.domElement=graph
        elif issubclass(graph.__class__,Graph.Graph) and type(name) in types.StringTypes:
            # initialise dom element
            self.domElement=xml.dom.minidom.Element(xmlGraphElement.graphElementName)
            self.setName(name)
            self.setGraph(graph)
        else:
            raise FileException("wrong arguments provided")
            
    def getAboutAsString(self):
        requestedElements=filter(lambda e:e.nodeType==xml.dom.Node.ELEMENT_NODE and e.tagName=="about",
                                 self.domElement.childNodes)
        if len(requestedElements)<1:
            return None
        else:
            return requestedElements[0].toxml()
            
    def getGraph(self):
        """
        returns the Graph object constructed  from this dom element
        """
        return GraphUtil.OpenCATBoxGraph(self.getGraphAsStringIO())
        
    def getGraphAsStringIO(self):
        text=StringIO.StringIO()
        for e in self.domElement.childNodes:
            if (e.nodeType==xml.dom.minidom.Node.TEXT_NODE or
                e.nodeType==xml.dom.minidom.Node.CDATA_SECTION_NODE):
                text.write(xmlDecode(e.data))
        text.seek(0)
        return text
        
    def setGraph(self,graph):
        # we were called with a Graph and a name as arguments
        data=StringIO.StringIO()
        GraphUtil.SaveCATBoxGraph(graph,data)
        #remove all other text nodes
        textNodes=filter(lambda e:e.nodeType==xml.dom.minidom.Node.TEXT_NODE or
                         e.nodeType==xml.dom.minidom.Node.CDATA_SECTION_NODE,
                         self.domElement.childNodes)
        map(self.domElement.removeChild,textNodes)
        # create necessary dom elements
        # maybe xml.dom.minidom.Text is ok, too
        newCDATASection=None
        if self.domElement.ownerDocument:
            newCDATASection=self.domElement.ownerDocument.createCDATASection(
                xmlEncode(data.getvalue())
                )
        else:
            newCDATASection=xml.dom.minidom.CDATASection(xmlEncode(data.getvalue()))
        self.domElement.appendChild(newCDATASection)
        
    def setGraphFromCATFile(self,GraphFile):
        """
        creates a graph object from .cat file and sets it
        this detour assures an actual graph format
        """
        self.setGraph(GraphUtil.OpenCATBoxGraph(GraphFile))
        
    def getName(self):
        return xmlDecode(self.domElement.getAttribute("name"))
        
    def setName(self,name):
        self.domElement.setAttribute("name",xmlEncode(name))
        
class xmlGatoElement:
    """
    is an access and modification interface to the gato dom structure 
    """
    
    gatoElementName="gato"
    
    def __init__(self,_domElement):
        """
        creates the access and modification interface
        """
        # test for right tag
        if _domElement.tagName!=self.gatoElementName:
            raise FileException("Wrong element name: %s, %s expected"
                            %(_domElement.tagName,self.gatoElementName))
        self.domElement=_domElement
        
    def createGraphElement(self,name=None):
        """
        creates a new graph Element
        """
        newDOMElement=None
        if self.domElement.ownerDocument:
            newDOMElement=self.domElement.ownerDocument.createElement(
                xmlGraphElement.graphElementName)
        else:
            newDOMElement=xml.dom.Element(xmlGraphElement.graphElementName)
        newElement=xmlGraphElement(newDOMElement)
        if name is not None:
            newElement.setName(name)
        return newElement
        
    def createAlgorithmElement(self,name=None):
        """
        creates a new algorithm element
        """
        newDOMElement=None
        if self.domElement.ownerDocument:
            newDOMElement=self.domElement.ownerDocument.createElement(
                xmlAlgorithmElement.algorithmElementName)
        else:
            newDOMElement=xml.dom.Element(xmlAlgorithmElement.algorithmElementName)
        newElement=xmlAlgorithmElement(newDOMElement)
        if name is not None:
            newElement.setName(name)
        return newElement
        
    def setName(self,name):
        """
        sets the name attribute for this element
        """
        self.domElement.setAttribute("name",xmlEncode(name))
        
    def getName(self,name):
        """
        gets the name of this element, if unset an empty string
        """
        return xmlDecode(self.domElement.getAttribute("name"))
        
    def getAboutAsString(self):
        requestedElements=filter(lambda e:e.nodeType==xml.dom.Node.ELEMENT_NODE and e.tagName=="about",
                                 self.domElement.childNodes)
        if len(requestedElements)<1:
            return None
        else:
            return requestedElements[0].toxml()
            
    def getGraphElements(self):
        """
        returns all direct child elements with graph's tagName
        """
        return filter(lambda x:(x.nodeType==xml.dom.minidom.Node.ELEMENT_NODE and
                                x.tagName==xmlGraphElement.graphElementName),
                      self.domElement.childNodes)
        
    def getAlgorithmElements(self):
        """
        returns all direct child elements with algorithm's tagName
        """
        return filter(lambda x:(x.nodeType==xml.dom.minidom.Node.ELEMENT_NODE and
                                x.tagName==xmlAlgorithmElement.algorithmElementName),
                      self.domElement.childNodes)
        
    def getGraphNames(self):
        """
        extracts all graph names
        """
        # return a list of name attribute values from graph elements
        return map(lambda x:xmlDecode(x.getAttribute("name")),
                   self.getGraphElements())
        
    def getAlgorithmNames(self):
        """
        extracts all algorithm names
        """
        # return a list of name attribute values from graph elements
        return map(lambda x:xmlDecode(x.getAttribute("name")),
                   self.getAlgorithmElements())
        
    def getGraphByName(self,name):
        """
        gets a xmlGraphElement by its name
        """
        graphs=filter(lambda x:x.getAttribute("name")==name,
                      self.getGraphElements())
        return xmlGraphElement(graphs[0])
        
    def getAlgorithmByName(self,name):
        """
        gets a xmlAlgorithmElement by its name
        """
        graphs=filter(lambda x:x.getAttribute("name")==name,
                      self.getAlgorithmElements())
        return xmlAlgorithmElement(graphs[0])
        
    def getDefaultSelection(self):
        """
        searches for the default selection, e.g.
        only one graph and one algorithm
        """
        defaultSelection={}
        graphs=self.getGraphNames()
        algorithms=self.getAlgorithmNames()
        if len(graphs)<2 and len(algorithms)<2:
            if len(graphs)==1:
                defaultSelection["graph"]=self.getGraphByName(graphs[0])
            if len(algorithms)==1:
                defaultSelection["algorithm"]=self.getAlgorithmByName(algorithms[0])
            return defaultSelection
            
        if self.domElement.hasAttribute("defaultGraph"):
            defaultSelection["graph"]=self.getGraphByName(
                self.domElement.getAttribute("defaultGraph"))
            
        if self.domElement.hasAttribute("defaultAlgorithm"):
            defaultSelection["algorithm"]=self.getAlgorithmByName(
                self.domElement.getAttribute("defaultAlgorithm"))
            
        return defaultSelection
        
    def updateGraphByName(self,graph,name=None):
        """
        appends or replaces a graph, if name not given, the graphs name is taken
        """
        newGraph=None
        newName=None
        if issubclass(graph.__class__,Graph.Graph):
            newGraph=xmlGraphElement(graph,name)
        elif issubclass(graph.__class__,xmlGraphElement):
            newGraph=graph
            if name:
                graph.setName(name)
            else:
                name=graph.getName()
        else:
            raise FileException("argument mismatch!")
            
        graphs=filter(lambda x:x.getAttribute("name")==name,
                      self.getGraphElements())
        if len(graphs):
            # replace graph, more exactly the first one
            self.domElement.replaceChild(newGraph.domElement,graphs[0])
        else:
            # append graph
            self.domElement.appendChild(newGraph.domElement)
            
    def updateAlgorithmByName(self,algorithm,name=None):
        """
        appends or replaces an algorithm
        """
        newAlgorithm=None
        newName=None
        if issubclass(algorithm.__class__,Gato.Algorithm):
            newAlgorithm=xmlAlgorithmElement(algorithm,name)
        elif issubclass(algorithm.__class__,xmlAlgorithmElement):
            newAlgorithm=algorithm
            if name:
                algorithm.setName(name)
            else:
                name=algorithm.getName()
        else:
            raise FileException("argument mismatch!")
            
        algorithms=filter(lambda x:x.getAttribute("name")==name,
                          self.getAlgorithmElements())
        if len(algorithms):
            # replace graph, more exactly the first one
            self.domElement.replaceChild(newAlgorithm.domElement,graphs[0])
        else:
            # append graph
            self.domElement.appendChild(newAlgorithm.domElement)
            
    def removeGraphByName(self,name):
        """
        removes the named graph
        """
        graphs=filter(lambda x:x.getAttribute("name")==name,
                      self.getGraphElements())
        map(self.domElement.removeChild,graphs)
        
    def removeAlgorithmByName(self,name):
        """
        removes the named algorithm
        """
        algorithms=filter(lambda x:x.getAttribute("name")==name,
                          self.getAlgorithmElements())
        map(self.domElement.removeChild,algorithms)
        
class ElementDisplay(Tkinter.Frame):
    """
    displays the selected element
    """
    def __init__(self,master,**config):
        """
        """
        config.update({"width":500, "height":500})
        Tkinter.Frame.__init__(self,master,bg="white",cnf=config)
        self.pack_propagate(0) # keep everything in a rigid frame
        self.displayedWidget=None
        self.displayedNode=None
        
    def clearDisplay(self):
        """
        cleanup and display nothing
        """
        # cleanup radically
        for child in self.children.values():
            child.pack_forget()
            child.destroy()
        self.displayedNode=None
        
    def setDisplay(self,node):
        """
        display node...
        """
        if node is None:
            self.clearDisplay()
            return
        else:
            if self.displayedNode is not None and self.displayedNode==node:
                return #nothing to do
        self.clearDisplay()
        displayedWidget=None
        if node.tagName==xmlGraphElement.graphElementName:
            html_data=xmlGraphElement(node).getAboutAsString()
            displayedWidget=AboutWidget(self,html_data)
            self.displayedNode=node
        elif node.tagName==xmlAlgorithmElement.algorithmElementName:
            displayedWidget=AlgorithmInfoWidget(self,xmlAlgorithmElement(node))
            self.displayedNode=node
        elif node.tagName==xmlGatoElement.gatoElementName:
            html_data=xmlGatoElement(node).getAboutAsString()
            displayedWidget=AboutWidget(self,html_data)
            self.displayedNode=node
        else:
            print "toDo: ",node
            
        if displayedWidget:
            displayedWidget.pack(expand=1,fill=Tkinter.BOTH)
            
            
class AlgorithmInfoWidget(Tkinter.Frame):
    """
    creates an algorithm widget
    """
    
    def __init__(self,master,algorithmElement):
        """
        create it
        """
        Tkinter.Frame.__init__(self,master,highlightthickness=0)
        html_data=algorithmElement.getAboutAsString()
        self.About=AboutWidget(self,html_data)
        self.About.pack(side=Tkinter.TOP,expand=1,fill=Tkinter.BOTH)
        text_data=algorithmElement.getText()
        self.Text=ScrolledText.ScrolledText(self,
                                            background='white', 
                                            wrap='word',
                                            font="Times 10")
        self.Text.insert(Tkinter.END, text_data)
        self.Text.pack(side=Tkinter.TOP,expand=1,fill=Tkinter.BOTH)
        self.Text['state'] = Tkinter.DISABLED 
        
class AboutWidget(ScrolledText.ScrolledText):
    """
    displays html text
    copied and modified from GatoDialogs.py
    Basic class which provides a scrollable area for viewing HTML text
    """
    
    def __init__(self, master, htmlcode):
        ScrolledText.ScrolledText.__init__(self, master,
                                           background='white', 
                                           wrap='word',
                                           font="Times 10",
                                           )
        self.inserthtml(htmlcode)
        
    def Update(self,htmlcode):
        self.inserthtml(htmlcode)
        
    def inserthtml(self, htmlcode):
        if htmlcode is None:
            htmlcode="""<html><body>
            No information avaliable !
            </body></html>
            """
        import formatter
        import GatoDialogs
        self['state'] = Tkinter.NORMAL
        self.delete('0.0', Tkinter.END)
        
        writer = GatoDialogs.HTMLWriter(self, self)
        format = formatter.AbstractFormatter(writer)
        parser = GatoDialogs.MyHTMLParser(format, self)
        parser.feed(htmlcode)
        parser.close()
        
        self['state'] = Tkinter.DISABLED
        
class GatoFileButtonBar(Tkinter.Frame):
    """
    holds the buttons for the node selection dialog
    """
    def __init__(self, master, reporter, cnf={}, **config):
    
        cnf.update(config)
        Tkinter.Frame.__init__(self,master,cnf=cnf)
        self.reporter=reporter
        self.okButton=Tkinter.Button(self,text="Ok",command=lambda e=None :self.reporter("quit"))
        self.okButton.grid(row=0,column=0,sticky=Tkinter.E+Tkinter.W)
        self.cancelButton=Tkinter.Button(self,text="Cancel",command=lambda e=None:self.reporter("cancel"))
        self.cancelButton.grid(row=0,column=1,sticky=Tkinter.E+Tkinter.W)
        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        
        
class GatoFileStructureWidget(TextTreeWidget.dom_structure_widget):
    """
    customized for .gato files
    """
    
    def __init__(self,master,dom,report_function,**conf):
        """
        report function coordinates the element display
        """
        TextTreeWidget.dom_structure_widget.__init__(self,master,dom,self.report_func,width=30,cnf=conf)
        self.chosenNodes={}
        self.report_to=report_function
        
    def report_func(self,event,*opts):
        """
        reports nodes that are marked aka chosen
        """
        if event=="chosenNode":
            # double click on node, or hit <Return>
            if len(opts)<1 or opts[0] is None:
                return
            node=opts[0]
            # is this a gato element?
            if node.tagName==xmlGatoElement.gatoElementName:
                if self.report_to and not self.report_to(event,node):
                    return
                self.chosenNode(node,xmlGatoElement.gatoElementName)
                # is it a valid graph or algorithm object?
            if self.isGatoMemberElement(node)!=1:
                # nothing to do...
                return
            if node.tagName==xmlAlgorithmElement.algorithmElementName:
                if self.report_to and not self.report_to(event,node):
                    return
                self.chosenNode(node,xmlAlgorithmElement.algorithmElementName)
            elif node.tagName==xmlGraphElement.graphElementName:
                if self.report_to and not self.report_to(event,node):
                    return
                self.chosenNode(node,xmlGraphElement.graphElementName)
        elif event=="selectedNode":
            # selected Node
            if len(opts)<1 or opts[0] is None:
                return
            node=opts[0]
            if self.report_to:
                self.report_to(event, node)
        else:
            print event, opts
            
    def chosenNode(self,node,register):
        """
        mark and save chosen node
        """
        last=self.chosenNodes.get(register)
        if last:
            self.setNodeColor(last,"black")
        if last==node:
            # erase selection
            self.chosenNodes[register]=None
        else:
            # new selection
            self.chosenNodes[register]=node
            self.setNodeColor(node,"red")
            
    def nameNode(self,node):
        """
        chose proper names for node
        """
        if node.nodeType==xml.dom.Node.ELEMENT_NODE:
            myName=xmlDecode(node.tagName)
            if myName==xmlGatoElement.gatoElementName:
                if node.hasAttribute("name"):
                    return "gato: %s"%xmlDecode(node.getAttribute("name"))
                else:
                    return "gato"
            elif myName==xmlGraphElement.graphElementName:
                return "graph: %s"%xmlDecode(node.getAttribute("name"))
            elif myName==xmlAlgorithmElement.algorithmElementName:
                return "algorithm: %s"%xmlDecode(node.getAttribute("name"))
            else:
                return TextTreeWidget.dom_structure_widget.nameNode(self,node)
        else:
            return TextTreeWidget.dom_structure_widget.nameNode(self,node)
            
    def isGatoMemberElement(self,node):
        """
        if this is a subtag of gato, return actual depth
        returns 0, if not in a gato element, or is a gato element itself
        """
        parent=node.parentNode
        if parent:
            if parent.nodeType==xml.dom.Node.DOCUMENT_NODE:
                return 0
            elif parent.nodeType==xml.dom.Node.ELEMENT_NODE and parent.tagName=="gato":
                return 1
            else:
                depth=self.isGatoMemberElement(parent)
                if depth:
                    return depth+1
                else:
                    return 0
        else:
            raise FileException("could not determine, if node is member of gato section")
            
    def isVisible(self,node):
        """
        filter out subsections of graph and algorithm in gato sections
        """
        isMember=self.isGatoMemberElement(node)
        if isMember==0:
            return TextTreeWidget.dom_structure_widget.isVisible(self,node)
        elif isMember>1:
            # hide everything deeper than 1
            return 0
        else:
            # hide everything except graph and algorithm
            if (node.nodeType==xml.dom.Node.ELEMENT_NODE and
                node.tagName in [xmlGraphElement.graphElementName,
                                 xmlAlgorithmElement.algorithmElementName]):
                return 1
            else:
                return 0
                
class GatoDOMDialog(Tkinter.Toplevel):
    """
    contains the xml structure and the display of selected element and some buttons
    opt_choose contains the elements, that may be chosen
    mand_choose contains the elements, that must be choosen
    """
    def __init__(self, master, dom, opt_choose=("algorithm","graph"), mand_choose=(), **config):
        """
        create it
        """
        self.dom=dom
        self.opt_choose=opt_choose
        self.mand_choose=mand_choose
        self.result={} # None in case of canceled dialog, (graphNode, algoNode) in case of selection
        Tkinter.Toplevel.__init__(self,master, cnf=config)
        # the structure widget...
        self.struct_widget=GatoFileStructureWidget(self, self.dom, self.reportFromStructure, font="Times 12")
        self.struct_widget.grid(row=0, column=0, sticky=Tkinter.NSEW)
        # and the element display....
        self.display_widget=ElementDisplay(self,width=400)
        self.display_widget.grid(row=0, column=1, sticky=Tkinter.NSEW)
        # the button bar...
        self.buttons=GatoFileButtonBar(self,self.reportFromButtons)
        self.buttons.grid(row=1, column=0, columnspan=2, sticky=Tkinter.EW)
        self.protocol("WM_DELETE_WINDOW", lambda e=None: self.reportFromButtons("cancel"))
        
    def reportFromStructure(self,event,node):
        """
        coordinates element display
        """
        if event=="selectedNode":
            self.display_widget.setDisplay(node)
        elif event=="chosenNode":
            # should be selected?
            if node and node.nodeType==xml.dom.Node.ELEMENT_NODE and \
               (node.tagName in self.opt_choose or node.tagName in self.mand_choose):
                return 1
                
    def reportFromButtons(self,event):
        """
        gets interesting events from button bar
        """
        if event=="quit":
            # graph
            graph=self.struct_widget.chosenNodes.get(xmlGraphElement.graphElementName)
            if graph:
                self.result[xmlGraphElement.graphElementName]=xmlGraphElement(graph)
            else:
                if xmlGraphElement.graphElementName in self.mand_choose:
                    return
                    # algorithm
            algorithm=self.struct_widget.chosenNodes.get(xmlAlgorithmElement.algorithmElementName)
            if algorithm:
                self.result[xmlAlgorithmElement.algorithmElementName]=xmlAlgorithmElement(algorithm)
            else:
                if xmlAlgorithmElement.algorithmElementName in self.mand_choose:
                    return
                    # gato section
            gato=self.struct_widget.chosenNodes.get(xmlGatoElement.gatoElementName)
            if gato:
                self.result[xmlGatoElement.gatoElementName]=xmlGatoElement(gato)
            else:
                if xmlGatoElement.gatoElementName in self.mand_choose:
                    return
            self.destroy()
            self.quit()
        elif event=="cancel":
            self.result=None
            self.destroy()
            self.quit()
        else:
            raise Exception("unknown event %s"%event)
            
    def run(self):
        """
        runs this dialog
        """
        self.mainloop()
        return self.result
        
class GatoFile:
    """
    encapsulates the dom creation from file
    """
    
    validModes=['r', # expect valid gato file and read this file
                'w', # create new file or overwrite existing file
                ]
    
    def __init__(self,myFile=None, mode='r'):
        """
        open a file and read the dom from it, or provide an empty gato document
        """
        if myFile:
            if mode not in self.validModes:
                raise FileException("mode %c not supported"%mode)
            self.mode=mode
            import types
            
            # test if name or file is given
            if self.mode=='r':
                self.file=None
                self.name=None
                if type(myFile) in types.StringTypes:
                    if os.access(myFile,os.R_OK):
                        self.file=file(myFile,"r")
                        self.name=myFile
                    else:
                        urlRequest=urllib2.Request(myFile)
                        self.file=urllib2.urlopen(urlRequest,"r")
                elif type(myFile)==types.FileType:
                    self.file=myFile
                    
                try:
                    # pull dom out of file
                    self.dom = xml.dom.minidom.parse(self.file)
                except xml.dom.DOMException, e:
                    print e
                    raise FileException(str(e))
                    
            elif self.mode=='w':
                if type(myFile) in types.StringTypes:
                    if os.access(myFile,os.W_OK):
                        self.file=file(myFile,"w")
                        self.name=myFile
                    else:
                        urlRequest=urllib2.Request(myFile)
                        self.file=urllib2.urlopen(urlRequest,"w")
                else:
                    self.file=myFile
                    
                    # create empty document
                self.dom = xml.dom.minidom.Document()
                
        else: #myFile=None
            # without file, but create empty doc, too
            self.dom = xml.dom.minidom.Document()
            # self.dom.appendChild(self.dom.createElement(xmlGatoElement.gatoElementName))
            
    def write(self,_file=None):
        """
        writes the (modified) dom to the file
        """
        # ???? write to right place????
        _thisFile=None
        if _file is not None:
            if type(_file) in types.StringTypes:
                _thisFile=encodedStream(file(_file,"w"))
            elif type(_file)==types.FileType:
                _thisFile=_file
        else:
            _thisFile=self.file
            
        encoding="iso-8859-1"
        _thisFile.write("<?xml version=\"1.0\" encoding=\"%s\" ?>\n"%encoding)
        if self.dom.hasChildNodes():
            for n in self.dom.childNodes:
                n.writexml(xmlEncodedStream(_thisFile))
                
    def getDefaultSelection(self):
        """
        returns the default selected elements
        """
        if (not self.dom.documentElement) or \
               self.dom.documentElement.tagName!=xmlGatoElement.gatoElementName:
            return None
        else:
            return xmlGatoElement(self.dom.documentElement).getDefaultSelection()
            
    def getGatoElements(self):
        """
        returns an xmlGatoElement instance
        """
        return self.dom.getElementsByTagName(xmlGatoElement.gatoElementName)
        
    def appendGatoElement(self,newElement):
        """
        append the element without checking unique name
        """
        if self.dom.hasChildNodes():
            # replace doc root gato node with a more general section
            if self.dom.documentElement.tagName==xmlGatoElement.gatoElementName:
                tmpGatoElement=self.dom.documentElement
                self.dom.replaceChild(self.dom.createElement("xml"),tmpGatoElement)
                self.dom.documentElement.appendChild(tmpGatoElement)
                # append to document element
            self.dom.documentElement.appendChild(newElement.domElement)
        else:
            # start with a gato node as root element
            self.dom.appendChild(newElement.domElement)
            
    def createGatoElement(self,name=None):
        """
        creates a new (opt. unnamed)  Gato element
        """
        newElement=xmlGatoElement(self.dom.createElement(xmlGatoElement.gatoElementName))
        if name is not None:
            newElement.setName(name)
        return newElement
        
    def displaySelectionDialog(self,master):
        """
        runs the dom structure
        """
        w=GatoDOMDialog(master,self.dom)
        return w.run()
        
    def displayGraphSelectionDialog(self,master):
        w=GatoDOMDialog(master,self.dom,to_choose=(('graph',)))
        s=w.run()
        if s:
            return s["graph"]
        else:
            return None
            
    def displayAlgorithmSelectionDialog(self,master):
        w=GatoDOMDialog(master, self.dom,to_choose=(('algorithm',)))
        s=w.run()
        if s:
            return s["algorithm"]
        else:
            return None
            
if __name__=='__main__':
    # some test progs
    WorkInProgress().run()
    
    # old ones
    # f=GatoFile('DFS.gato') # parse an XML file by name
    # f.displayGraphSelectionDialog(Tkinter.Tk())
    
    
